{-# OPTIONS_GHC -Wno-orphans #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE DataKinds #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE RankNTypes #-}
module Math.Alspach where

import Linear.V
import qualified Data.Vector as DV
import GHC.TypeNats hiding (Mod)
import GHC.TypeLits.Witnesses
import Data.Type.Equality
import Data.Proxy
import Data.Functor.Rep
import Data.Foldable
import Control.Monad
import Math.NumberTheory.Primes.Factorisation
import Math.NumberTheory.Moduli
import Data.Text.Prettyprint.Doc

import Math.Polynomial

-- This is a bit ugly!
succKN :: forall n r. KnownNat n => Proxy n -> (KnownNat (n + 1) => r) -> r
succKN _ x = case SNat @n %+ SNat @1 of SNat -> x

instance KnownNat n => Pretty (Mod n) where pretty = viaShow

mBound :: forall n. KnownNat n => Proxy n -> Monomial n
mBound prox = constExpM (n - 1)
  where n = reflectDim prox

termBounded :: KnownNat n => Proxy n -> Term n a -> Bool
termBounded prox (T (m, _)) = m <<= mBound prox

-- This is the type of the function that computes the bounded terms of the
-- conjecture polynomial. I'm defining it as a type synonym because we have
-- multiple different implementations of that function.
type TermFunction = forall n a.
  (KnownNat n, Num a, Ord a) => Proxy n -> [Term n a]

{- Version #1. Define the polynomial, so that we can then filter out its
 - maximal and bounded terms, using a polynomial definition based on the Coq.
 - Slowest. -}

partialSums :: (KnownNat (n + 1), Num a) => V n a -> V (n + 1) a
partialSums v = tabulate $ \i -> sum (take i (toList v))

ndistinctPolyFactors :: forall n a. (KnownNat n, Num a, Eq a) =>
  (Int -> Int -> Bool) -> [Polynomial n a]
ndistinctPolyFactors p =
  [var i - var j | i <- [0..n - 1], j <- [0..n - 1], p i j]
  where n = reflectDim @n Proxy

ndistinctPoly ::
  (KnownNat n, Num a, Eq a) => (Int -> Int -> Bool) -> Polynomial n a
ndistinctPoly p = product (ndistinctPolyFactors p)

psumsNdistinctPoly :: forall n a. (KnownNat n, Num a, Eq a) => Polynomial n a
psumsNdistinctPoly = succKN @n Proxy $
  subst (ndistinctPoly @(n + 1) (\i j -> i < j && not (exc i j)))
        (partialSums (tabulate var))
  where n = reflectDim @n Proxy
        exc i j = (i == 0 && j == n) || j == i + 1

nAlspachInstPoly :: (KnownNat n, Num a, Eq a) => Polynomial n a
nAlspachInstPoly = ndistinctPoly (<) * psumsNdistinctPoly

candidateTerms :: TermFunction
candidateTerms prox = filter (termBounded prox) . terms $ nAlspachInstPoly

{- Version #2. Same approach, but the definition of the polynomial is based on
 - the one in the Hicks-Ollis-Schmitt paper. Faster. -}

nAlspachInstPoly' :: forall n a.
  (KnownNat n, Num a, Eq a) => Polynomial n a
nAlspachInstPoly' = product [
  (var j - var i) * if (i, j) /= (0, n - 1)
                       then sum (map var [i..j]) else 1 |
  i <- [0..n - 1], j <- [i + 1..n - 1]]
  where n = reflectDim @n Proxy

candidateTerms' :: TermFunction
candidateTerms' prox = filter (termBounded prox) . terms $ nAlspachInstPoly'

{- Version #3. Build a list of factors, based on the Coq-inspired definition,
 - and do a bounded multiplication. Even faster. -}

psumsNdistinctPolyFactors :: forall n a.
  (KnownNat n, Num a, Eq a) => [Polynomial n a]
psumsNdistinctPolyFactors = [-sum (map var [i..j - 1]) |
  -- filter removes cases already handled by preconditions of conjecture
  i <- [0..n], j <- [0..n], i < j, not (i == 0 && j == n), j /= i + 1]
  where n = reflectDim @n Proxy

nAlspachInstPolyFactors ::
  (KnownNat n, Num a, Eq a) => [Polynomial n a]
nAlspachInstPolyFactors =
  ndistinctPolyFactors (<) ++ psumsNdistinctPolyFactors

candidateTermsViaBP :: TermFunction
candidateTermsViaBP prox =
  terms $ boundedProduct (mBound prox) nAlspachInstPolyFactors

{- Version #4. Build a grouping of factors, based on the HOM paper's
 - definition, and do a bounded multiplication. Fastest... so far. -}

nAlspachInstPolyFactors' :: forall n a.
  (KnownNat n, Num a, Eq a) => Expr n a
nAlspachInstPolyFactors' = join . groupR (Leaf 1) $ [
  Leaf (var j - var i) `Prod`
  Leaf (if (i, j) /= (0, n - 1) then sum (map var [i..j]) else 1) |
  i <- [0..n - 1], j <- [i + 1..n - 1]]
  where n = reflectDim @n Proxy

candidateTermsViaBP' :: TermFunction
candidateTermsViaBP' prox =
  terms $ boundedProductE (mBound prox) nAlspachInstPolyFactors'

candidateTermsViaMSBP' :: TermFunction
candidateTermsViaMSBP' prox = terms $
  boundedProductViaMS minmax (mBound prox) (toList nAlspachInstPolyFactors')


factorsStep :: forall n a. (KnownNat n, Ord a, Num a) =>
  Maybe (Expr (n + 1) a)
factorsStep = succKN @n Proxy $
  divideReplace 1 (widen1E nAlspachInstPolyFactors')
    nAlspachInstPolyFactors'

-- Just (candidateTermsViaBP' (Proxy @6)) ==
-- (terms . boundedProductE (mBound Proxy) <$> factored)
factored :: (KnownNat n, Ord a, Num a) =>
  Maybe (Expr (n + 1) a)
factored = (widen1E nAlspachInstPolyFactors' `Prod`) <$> factorsStep

factorsStepBP :: forall n a. (KnownNat n, Ord a, Num a) =>
  Maybe (Polynomial (n + 1) a)
factorsStepBP = succKN @n Proxy $
  let bound = mBound Proxy -<> widen1M (mBound Proxy)
  in boundedProductE bound <$> factorsStep


type QuotSeq (n :: Nat) t = V n (Mod t)
type QSType t = Mod t -> Int

nCompositeAlspachInstPolyFactorsP :: forall n t.
  (KnownNat n, KnownNat (n + 1), KnownNat t) =>
  QuotSeq n t -> Expr n Integer
nCompositeAlspachInstPolyFactorsP a = join . groupR (Leaf 1) $ [
  gif (index a i == index a j) (var j - var i) `Prod`
  {- (i, j) /= (0, n - 1) && ? -}
  gif ((i, j) /= (0, n - 1) && index b i == index b (j + 1))
      (sum (map var [i..j])) |
  i <- [0..n - 1], j <- [i + 1..n - 1]]
  where b = partialSums a
        gif True = Leaf
        gif False = const (Leaf 1)
        n = reflectDim @n Proxy

pBound :: forall n t. (KnownNat n) => QuotSeq n t -> Monomial n
pBound a = Monomial ((\n -> length (filter (== n) (toList a)) - 1) <$> a)

candidateTermsViaBPComposite :: forall n t.
  (KnownNat n, KnownNat (n + 1), KnownNat t) =>
  QuotSeq n t -> [Term n Integer]
candidateTermsViaBPComposite a =
  terms $ boundedProductE (pBound a) (nCompositeAlspachInstPolyFactorsP a)

candidateTermsViaMSBPComposite :: forall n t.
  (KnownNat n, KnownNat (n + 1), KnownNat t) =>
  QuotSeq n t -> [Term n Integer]
candidateTermsViaMSBPComposite a = terms $
  boundedProductViaMS minmax (pBound a)
    (toList (nCompositeAlspachInstPolyFactorsP a))

twoseq :: (Int, Int) -> (forall n. KnownNat n => QuotSeq n 2 -> r) -> r
twoseq (zeroes, ones0) = reifyVectorNat v
  where (ones, oneslast) = if even ones0 && ones0 > 0
                              then (ones0 - 1, 1) else (ones0, 0)
        z1 = zeroes `div` 2
        z2 = zeroes - z1
        v = foldMap (uncurry DV.replicate)
          [(z1, 0), (ones, 1), (z2, 0), (oneslast, 1)]

-- Int arg is how many zeroes we want.
twoseq' :: forall n. KnownNat n => Proxy n -> Int -> QuotSeq n 2
twoseq' prox zeroes = twoseq (zeroes', ones) recover
  where n = reflectDim prox
        zeroes' = max 0 (min zeroes n)
        ones = n - zeroes'
        recover :: forall n'. KnownNat n' => QuotSeq n' 2 -> QuotSeq n 2
        recover v = case sameNat (Proxy @n') (Proxy @n) of
          Just Refl -> v
          Nothing -> error "bad vector length in twoseq'?!"


-- Stuff that can be defined in terms of one of the above implementations.

candidateCoeffs :: KnownNat n => (p -> [Term n Integer]) -> p -> [Integer]
candidateCoeffs tf prox = map getCoeff (tf prox)

failedChars :: KnownNat n => (p -> [Term n Integer]) -> p -> Maybe [Integer]
failedChars tf prox = case candidateCoeffs tf prox of
    [] -> Nothing
    coeffs -> Just . map fst .  factorise . foldl' gcd 0 $ coeffs

failedPrimes :: forall n p.
  KnownNat n => (p -> [Term n Integer]) -> p -> Maybe [Integer]
{- Online friend points out that the sum of all nonzero elements in a prime
 - field is zero (except in F_2), so sets of any cardinality larger than *two*
 - less than the size of the field are out of scope. The F_2 case doesn't
 - matter, since we know the conjecture holds there anyway. -}
failedPrimes tf prox = filter (\char -> n <= char - 2) <$> failedChars tf prox
  where n = toInteger (reflectDim @n Proxy)


failedCharsTwo :: KnownNat n => Proxy n -> Maybe [Integer]
failedCharsTwo prox = concat <$> traverse go types
  where n = reflectDim prox
        types = [(zeroes, n - zeroes) | zeroes <- [0..n]]
        go lam = twoseq lam go'
        go' :: forall n'. KnownNat n' => QuotSeq n' 2 -> Maybe [Integer]
        go' = succKN (Proxy @n') (failedChars candidateTermsViaBPComposite)

failedPrimesTwo :: KnownNat n => Proxy n -> Maybe [Integer]
failedPrimesTwo prox =
  filter (\char -> n <= char * 2 - 1) <$> failedCharsTwo prox
  where n = toInteger (reflectDim prox)

