{-# OPTIONS_GHC -Wno-orphans #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE DataKinds #-}
{-# LANGUAGE DeriveTraversable #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE RankNTypes #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE FlexibleInstances #-}
module Math.Polynomial (
  -- * Working with this module
  -- $working

  -- * Why am I getting @No instance for (KnownNat n0)@ errors?!
  -- $watchout

  -- * Monomials
  Monomial(..),
  varM,
  constExpM,
  powM,
  (-<>),
  (<<=),
  mfactorings,
  degM,
  evalM,
  widen1M,
  (@@),

  -- * Polynomials
  Polynomial(..),
  Term(..),
  getM,
  getCoeff,
  terms,
  fromCoeffs,
  fromTerms,
  filterTerms,
  (*:),
  (*!),
  var,
  constant,
  eval,
  subst,
  Degree,
  neginf,
  deg,
  leading,
  homogeneous,
  widen1T,
  widen1,

  -- * Groupings
  Grouping(..),
  foldG,
  foldMapG,
  foldG',
  foldMapG',
  groupR,
  groupL,
  groupB,
  divide,
  divideReplace,

  -- * Groupings of polynomials
  Expr,
  homogeneousE,
  degE,
  widen1E,

  -- * Bounded polynomial multiplication
  -- $bounded
  reduce,
  boundedMul,
  boundedProductE,
  boundedProduct,

  -- * Multiset-based multiplication
  -- $multisets
  Factor(..),
  lexico,
  FactorMS,
  Selection(..),
  SelFun,
  boundedProductMS,
  boundedProductViaMS,
  minmax,
  ) where

import Linear.V
import Linear.Vector
import qualified Data.Vector as DV
import Data.Map.Strict (Map)
import qualified Data.Map.Strict as M
import Data.Maybe
import Data.Monoid
import Data.Functor.Rep
import Data.Foldable
import Control.Monad.Zip
import GHC.TypeNats
import GHC.TypeLits.Witnesses
import Control.Monad
import Control.Monad.State
import Data.Text.Prettyprint.Doc
import Data.String
import Data.Proxy
import Data.MultiSet (MultiSet)
import qualified Data.MultiSet as MS
import Data.Function
import Data.Ord

-- $setup
-- >>> :set -XDataKinds

{- $working
   == Extensions to enable
   You will almost certainly need to enable @DataKinds@ in order to use
   type-level 'Nat' literals, and you will likely also want at least
   @ScopedTypeVariables@ and @TypeApplications@.

   == Use pretty-printing!
   Most of the core types in this package have 'Pretty' instances, and most of
   these instances are /much/ nicer to look at than the 'Show' instances! When
   working in GHCi, be sure to apply 'pretty' to anything you want to look at.
   You may need to import 'Data.Text.Prettyprint.Doc' first. Compare:

   >>> var 3 :: Polynomial 5 Integer
   Polynomial {_getCoeffs = fromList [(Monomial {getExps = V {toVector = [0,0,0,1,0]}},1)]}
   >>> pretty (var 3 :: Polynomial 5 Integer)
   x₃
-}

{- $watchout
   This is the easiest stumbling block to hit when using this library! This
   misleading error message is really because of the phenomenon of
   /ambiguous instances/.

   == Background on ambiguous instances
   Ambiguous instances can show up in everyday Haskell as well, although it is
   less common. Consider an expression like

   > show 3 :: String

   It is not actually clear, here, what type the @3@ should have. It could be
   'Integer', 'Double', or any number of other types that have instances for
   both 'Num' and 'Show'—and the result of the expression depends on which
   one you mean!

   >>> show (3 :: Integer) == show (3 :: Double)
   False

   The basic issue is that a function which is polymorphic in its input type,
   @show@, is being applied to a term which is polymorphic in its result type,
   @3@, in such a way that the type being selected is no longer part of the
   overall result type, @String@, but the choice nonetheless makes a
   difference in behavior because of typeclasses. (This is distinct from a case
   like @length []@, which satisfies everything in the description except the
   final clause.)

   == Why ambiguous instances show up in this library
   This library makes extensive use of type-level 'Nat's to indicate the number
   of variables associated with an object, usually a 'Monomial' or
   'Polynomial'—for example, a polynomial in 3 variables with 'Integer'
   coefficients has type @Polynomial 3 Integer@. Furthermore, many or most
   functions are /not/ truly parametric in the number of variables; at some
   level, they may be making use of that information. In some cases, the
   non-parametricity is even visible:

   >>> pretty (var 10 :: Polynomial 3 Integer)
   1
   >>> pretty (var 10 :: Polynomial 20 Integer)
   x₁₀

   So most operations want to know how many variables something is in, and most
   operations do nothing to constrain that; the result is that type inference
   correctly recognizes that there is missing information.

   TODO: This is not quite a full explanation, and it doesn't really explain
   why this happens a lot in the REPL and less in the body of a function in a
   file. Need to finish writing this.

   == How to solve ambiguous instance issues
   Add some explicit type annotations in enough places that it's unambiguous
   how many variables every object you're talking about involves. For example:

   >>> pretty (var 3 + 5)
   <BLANKLINE>
   <interactive>:...: error:
       • No instance for (KnownNat n0) arising from a use of ‘+’
       • In the first argument of ‘pretty’, namely ‘(var 3 + 5)’
         In the expression: pretty (var 3 + 5)
         In an equation for ‘it’: it = pretty (var 3 + 5)
   >>> pretty (var 3 + 5 :: Polynomial 7 Integer)
   x₃ + 5
   >>> pretty ((var 3 :: Polynomial 7 Integer) + 5)
   x₃ + 5
   >>> pretty (var 3 + (5 :: Polynomial 7 Integer))
   x₃ + 5

   TODO: What if the user is getting an error like this in a polymorphic
   function? Should probably point to docu on @ScopedTypeVariables@.
-}


-- This is a bit ugly!
succKN :: forall n r. KnownNat n => Proxy n -> (KnownNat (n + 1) => r) -> r
succKN _ x = case SNat @n %+ SNat @1 of SNat -> x


-- | A @'Monomial' n@ is a monomial in @n@ variables, which we refer to as x₀
-- through xₙ₋₁. The exponents are stored as 'Int's rather than 'Integer's,
-- which should /hopefully/ not be a problem...
newtype Monomial (n :: Nat) =
  Monomial {getExps :: V n Int} deriving (Eq, Ord, Show, Read)

ppSup, ppSub :: Int -> String
ppSup = map perChar . show
  where perChar c = fromMaybe c (lookup c table)
        table = [('-', '¯'), ('0', '⁰'), ('1', '¹'), ('2', '²'), ('3', '³'),
                 ('4', '⁴'), ('5', '⁵'), ('6', '⁶'), ('7', '⁷'), ('8', '⁸'),
                 ('9', '⁹')]
ppSub = map perChar . show
  where perChar c = fromMaybe c (lookup c table)
        table = [('0', '₀'), ('1', '₁'), ('2', '₂'), ('3', '₃'), ('4', '₄'),
                 ('5', '₅'), ('6', '₆'), ('7', '₇'), ('8', '₈'), ('9', '₉')]

ppMonomial :: Monomial n -> String
ppMonomial (Monomial m) =
    case catMaybes (zipWith perVar [0..] (toList m)) of
      [] -> "1"
      vs -> unwords vs
  where perVar _ 0 = Nothing
        perVar sub 1 = Just $ "x" <> ppSub sub
        perVar sub sup = Just $ "x" <> ppSub sub <> ppSup sup
instance Pretty (Monomial n) where pretty = fromString . ppMonomial
-- instance Show (Monomial n) where show = ppMonomial

-- | Monomial multiplication is provided as the 'Semigroup' instance's
-- operation '<>'.
instance KnownNat n => Semigroup (Monomial n) where
  Monomial m1 <> Monomial m2 = Monomial (m1 ^+^ m2)
-- | 'mempty' is the constant 'Monomial', printed as @1@.
instance KnownNat n => Monoid (Monomial n) where
  mempty = Monomial zero

-- | A single variable as a 'Monomial'.
--
-- >>> pretty (varM 2 :: Monomial 4)
-- x₂
varM :: KnownNat n => Int -> Monomial n
varM i = Monomial v
  where v = tabulate (\j -> if j == i then 1 else 0)

-- | A 'Monomial' with the same exponent for each variable.
--
-- >>> pretty (constExpM 2 :: Monomial 6)
-- x₀² x₁² x₂² x₃² x₄² x₅²
constExpM :: KnownNat n => Int -> Monomial n
constExpM m = Monomial (tabulate (const m))

-- | Take a power of a 'Monomial'.
--
-- >>> pretty (powM (varM 2 <> varM 3 <> varM 3) 6 :: Monomial 5)
-- x₂⁶ x₃¹²
powM :: Monomial n -> Int -> Monomial n -- KnownNat n =>
powM (Monomial m) n = Monomial (m ^* n)

-- | Monomial division.
--
-- >>> pretty (constExpM 4 -<> varM 3 :: Monomial 5)
-- x₀⁴ x₁⁴ x₂⁴ x₃³ x₄⁴
-- >>> pretty (constExpM 4 -<> powM (varM 3) 10 :: Monomial 5)
-- x₀⁴ x₁⁴ x₂⁴ x₄⁴
(-<>) :: KnownNat n => Monomial n -> Monomial n -> Monomial n
Monomial m1 -<> Monomial m2 = Monomial (max 0 <$> m1 ^-^ m2)
infix 6 -<>

-- | The divisibility partial order on 'Monomial's.
--
-- >>> (varM 3 :: Monomial 5) <<= varM 3 <> varM 4
-- True
-- >>> (varM 3 :: Monomial 5) <<= varM 2
-- False
-- >>> (varM 2 :: Monomial 5) <<= varM 3
-- False
(<<=) :: KnownNat n => Monomial n -> Monomial n -> Bool
Monomial m1 <<= Monomial m2 = and (mzipWith (<=) m1 m2)
infix 4 <<=

-- | Find all of the factorizations of a 'Monomial' into two 'Monomial's.
-- We should have @(x, y) \`elem\` mfactorings m@ iff @x <> y == m@, and
-- @mfactorings m@ should contain no duplicates.
--
-- >>> pretty (mfactorings (constExpM 3 :: Monomial 2))
-- [ (x₀³ x₁³, 1)
-- , (x₀³ x₁², x₁)
-- , (x₀³ x₁, x₁²)
-- , (x₀³, x₁³)
-- , (x₀² x₁³, x₀)
-- , (x₀² x₁², x₀ x₁)
-- , (x₀² x₁, x₀ x₁²)
-- , (x₀², x₀ x₁³)
-- , (x₀ x₁³, x₀²)
-- , (x₀ x₁², x₀² x₁)
-- , (x₀ x₁, x₀² x₁²)
-- , (x₀, x₀² x₁³)
-- , (x₁³, x₀³)
-- , (x₁², x₀³ x₁)
-- , (x₁, x₀³ x₁²)
-- , (1, x₀³ x₁³) ]
mfactorings :: KnownNat n => Monomial n -> [(Monomial n, Monomial n)]
mfactorings m_@(Monomial m) = [(m_ -<> d, d) | d <- diffs]
  where diffs = Monomial <$> traverse (\n -> [0..n]) m

-- | The degree of a 'Monomial'—the sum of its exponents.
--
-- >>> pretty (constExpM 3 :: Monomial 5)
-- x₀³ x₁³ x₂³ x₃³ x₄³
-- >>> degM (constExpM 3 :: Monomial 5)
-- 15
degM :: Monomial n -> Int
degM (Monomial m) = sum m

-- | Substitute values for each of the variables in a 'Monomial' and evaluate
-- the result.
--
-- >>> m = varM 0 <> varM 0 <> varM 1 :: Monomial 2
-- >>> pretty m
-- x₀² x₁
-- >>> evalM m (toV (Linear.V2 3 4)) :: Integer
-- 36
evalM :: (KnownNat n, Num a) => Monomial n -> V n a -> a
evalM (Monomial m) v = product (mzipWith (^) v m)

-- | Convert a 'Monomial' in @n@ variables to a 'Monomial' in @n + 1@
-- variables that just happens to have an exponent of 0 for xₙ.
widen1M :: forall n. KnownNat n => Monomial n -> Monomial (n + 1)
widen1M (Monomial m) = Monomial m'
  where m' = succKN @n Proxy $
          fromMaybe (error "?!") . fromVector . (`DV.snoc` 0) . toVector $ m

-- | Get an exponent from a 'Monomial' by the index of the variable. This
-- function is partial! Be sure to pass an index less than @n@.
--
-- >>> (varM 3 <> varM 4 <> varM 3 :: Monomial 5) @@ 3
-- 2
(@@) :: KnownNat n => Monomial n -> Int -> Int
Monomial m @@ n = index m n


-- | A @'Polynomial' n a@ is a polynomial in @n@ variables with coefficients
-- from @a@. This is a polynomial /in expanded form/, not some kind of
-- unevaluated expression representing a polynomial, so ring operations on
-- 'Polynomial's can be expensive! See 'Grouping' if you want to defer stuff.
--
-- A 'Polynomial' is represented as a 'Map' from 'Monomial's to coefficients.
-- We keep an invariant of never having a key map to @0@, in order to minimize
-- the the size of maps and to avoid a distinction between a 'Monomial' being
-- "present with a coefficient of @0@" and being absent. Note that this means
-- that a nonzero constant 'Polynomial' is considered to have one term, while
-- the zero 'Polynomial' has none. In order to enforce this invariant, you
-- should generally use the 'fromCoeffs' function instead of the 'Polynomial'
-- data constructor.
--
-- Note that a variety of unexpected operations put 'Num' and/or 'Eq'
-- constraints on the coefficient type. This is usually to manage the
-- aforementioned invariant. In the case of pretty-printing, it's to detect
-- certain special cases.
newtype Polynomial (n :: Nat) a = Polynomial {_getCoeffs :: Map (Monomial n) a}
  deriving (Eq, Ord, Show, Read, Functor, Foldable, Traversable)
-- | A @'Term' n a@ is the same thing as a @('Monomial' n, a)@, but in a
-- newtype wrapper. We keep an invariant that anything in this newtype wrapper
-- has a nonzero second component. We think of a @'Polynomial'@ as a sum of
-- @'Term'@s.
newtype Term n a = T {unT :: (Monomial n, a)} deriving (Eq, Ord, Show, Functor)

-- | A list of the 'Term's in a 'Polynomial'.
--
-- >>> pretty (terms (7 * var 3 * var 2 + 2) :: [Term 4 Integer])
-- [2, 7 x₂ x₃]
terms :: Polynomial n a -> [Term n a]
terms = map T . M.assocs . _getCoeffs

-- | The 'Monomial' of a 'Term'.
--
-- >>> pretty (map getM (terms (7 * var 3 * var 2 + 2) :: [Term 4 Integer]))
-- [1, x₂ x₃]
getM :: Term n a -> Monomial n
getM = fst . unT

-- | The coefficient of a 'Term'.
--
-- >>> pretty (map getCoeff (terms (7 * var 3 * var 2 + 2) :: [Term 4 Integer]))
-- [2, 7]
getCoeff :: Term n a -> a
getCoeff = snd . unT

-- | Make a 'Polynomial' from an underlying 'Map', but filter out any spurious
-- mappings of 'Monomial's to @0@ first.
--
-- >>> m = M.fromList [(varM 3, 0), (varM 4 <> varM 3, 5), (mempty, 2)]
-- >>> pretty (fromCoeffs m :: Polynomial 5 Integer)
-- 5 x₃ x₄ + 2
fromCoeffs :: (Eq a, Num a) => Map (Monomial n) a -> Polynomial n a
fromCoeffs = Polynomial . M.filter (/= 0)

-- | Make a 'Polynomial' by adding up a list of 'Term's. "Adding up" here does
-- mean that if the same 'Monomial' appears more than once, the coefficients of
-- each of its appearances will be summed. This can potentially lead to
-- cancellation, meaning that not every 'Monomial' appearing in the input list
-- need appear in the output 'Polynomial'!
--
-- >>> ts1 = [T (varM 4 <> varM 3, 5), T (mempty, 2)]
-- >>> pretty (fromTerms ts1 :: Polynomial 5 Integer)
-- 5 x₃ x₄ + 2
-- >>> ts2 = [T (varM 1, 5), T (varM 1, 4), T (mempty, 1), T (mempty, -1)]
-- >>> pretty (fromTerms ts2 :: Polynomial 3 Integer)
-- 9 x₁
fromTerms :: (Eq a, Num a) => [Term n a] -> Polynomial n a
fromTerms = fromCoeffs . M.fromListWith (+) . map unT

-- | Filter the 'Term's of a 'Polynomial' by a predicate.
--
-- >>> p = var 3 + var 4 + var 5 * var 6 :: Polynomial 10 Integer
-- >>> pretty p
-- x₃ + x₄ + x₅ x₆
-- >>> pretty (filterTerms ((<= 1) . degM . getM) p)
-- x₃ + x₄
filterTerms :: (Term n a -> Bool) -> Polynomial n a -> Polynomial n a
filterTerms f = Polynomial . M.filterWithKey (curry (f . T)) . _getCoeffs

-- | Multiply a 'Polynomial' by a coefficient—i.e., the action of a ring on its
-- polynomial ring as a module over it.
--
-- >>> pretty (3 *: (var 3 + var 4 + var 5 * var 6 :: Polynomial 10 Integer))
-- 3 x₃ + 3 x₄ + 3 x₅ x₆
infix *:
(*:) :: (Eq a, Num a, KnownNat n) => a -> Polynomial n a -> Polynomial n a
0 *: _ = 0
coeff *: Polynomial p = Polynomial (fmap (coeff *) p)

-- | Multiply a 'Polynomial' by a 'Term'.
--
-- >>> pretty (T (varM 3, 3) *! (var 3 + var 4 + var 5 * var 6 :: Polynomial 10 Integer))
-- 3 x₃² + 3 x₃ x₄ + 3 x₃ x₅ x₆
infix *!
(*!) :: (Num a, KnownNat n) => Term n a -> Polynomial n a -> Polynomial n a
T (m, coeff) *! Polynomial p =
  Polynomial . M.mapKeys (m <>) . fmap (coeff *) $ p

ppTerm :: (Eq a, Num a) => (a -> Doc ann) -> Term n a -> Doc ann
ppTerm ppCoeff (T (m, coeff))
  | all (== 0) (getExps m) = ppCoeff coeff
  | otherwise = case coeff of
      1  -> pretty m
      -1 -> "-" <> pretty m
      _  -> ppCoeff coeff <+> pretty m
instance (Pretty a, Eq a, Num a) => Pretty (Term n a) where
  pretty = ppTerm pretty

ppPolynomial :: (Eq a, Num a) => (a -> Doc ann) -> Polynomial n a -> Doc ann
ppPolynomial ppCoeff (Polynomial p)
  | M.null p = "0"
  | otherwise = fillSep . punctuate " +" .
      map (ppTerm ppCoeff . T) . M.toDescList $ p
instance (Pretty a, Eq a, Num a) => Pretty (Polynomial n a) where
  pretty = ppPolynomial pretty
{-
instance (Show a, Eq a, Num a) => Show (Polynomial n a) where
  show = ppPolynomial
-}

-- | Standard ring operations on 'Polynomial's. 'signum' and 'abs' are not
-- implemented.
instance (KnownNat n, Num a, Eq a) => Num (Polynomial n a) where
  Polynomial p + Polynomial q = fromCoeffs (M.unionWith (+) p q)
  -- TODO maybe optimize this?
  p * q = fromCoeffs . M.unionsWith (+) $ [_getCoeffs (t *! p) | t <- terms q]
  signum = error "not actually defined for polynomials"
  abs = error "not actually defined for polynomials"
  fromInteger = constant . fromInteger
  negate (Polynomial p) = Polynomial (negate <$> p)

-- | A single variable as a 'Polynomial'.
--
-- >>> pretty (var 2 :: Polynomial 4 Integer)
-- x₂
var :: (KnownNat n, Num a) => Int -> Polynomial n a
var i = Polynomial (M.singleton (varM i) 1)

-- | A constant 'Polynomial'.
--
-- >>> pretty (constant 33 :: Polynomial 4 Integer)
-- 33
constant :: (KnownNat n, Num a, Eq a) => a -> Polynomial n a
constant 0 = Polynomial mempty
constant x = Polynomial (M.singleton mempty x)

-- | Substitute values for each of the variables in a 'Polynomial' and evaluate
-- the result.
--
-- >>> p = var 0 * var 1 + 1 :: Polynomial 2 Integer
-- >>> pretty p
-- x₀ x₁ + 1
-- >>> eval p (toV (Linear.V2 (-1) 2))
-- -1
eval :: (KnownNat n, Num a) => Polynomial n a -> V n a -> a
eval p v = sum [coeff * evalM m v | T (m, coeff) <- terms p]

-- | Substitute 'Polynomial's for each of the variables in another
-- 'Polynomial'. @subst p qs@ should be the same as
-- @eval (fmap constant p) qs@.
--
-- >>> p = var 0 + var 1 :: Polynomial 2 Integer
-- >>> pretty (subst p (toV (Linear.V2 (var 3 * var 4 :: Polynomial 7 Integer) 5)))
-- x₃ x₄ + 5
subst :: (KnownNat n, KnownNat m, Num a, Eq a) =>
  Polynomial n a -> V n (Polynomial m a) -> Polynomial m a
subst p v = sum [constant coeff * evalM m v | T (m, coeff) <- terms p]

-- | The type of degrees of 'Polynomial's. We think of it as consisting of
-- 'Int' augmented by a new element @-∞@ ('neginf'). We could just define this
-- as @'Maybe' 'Int'@; then the finite 'Int's would be the 'Just's, and @-∞@
-- would be represented by 'Nothing'. This even has the right 'Ord' instance!
-- However, by using some newtype wrappers from "Data.Monoid", we can further
-- get addition of degrees as the 'Monoid' instance for free, as well as a
-- (somewhat wonky) 'Num' instance.
--
-- We have an orphan 'Pretty' instance for 'Degree'.
--
-- >>> [4, neginf]
-- [Ap {getAp = Just (Sum {getSum = 4})},Ap {getAp = Nothing}]
-- >>> pretty [4, neginf]
-- [4, -∞]
-- >>> pretty [neginf <> 2, 3 <> 4]
-- [-∞, 7]
-- >>> (3 :: Degree) < 4
-- True
-- >>> neginf < 0
-- True
type Degree = Ap Maybe (Sum Int)
instance Pretty Degree where
  pretty (Ap Nothing) = "-∞"
  pretty (Ap (Just (Sum n))) = pretty n

-- | Convert an 'Int' to a finite 'Degree'.
findeg :: Int -> Degree
findeg = Ap . Just . Sum

-- | @-∞@, the 'Degree' of the @0@ 'Polynomial'.
neginf :: Degree
neginf = Ap Nothing

-- | The degree of a 'Polynomial'. The maximum 'degM' of its 'Monomial's, if it
-- has any, or 'neginf' for the @0@ 'Polynomial'.
--
-- >>> pretty (deg (var 0 :: Polynomial 1 Integer))
-- 1
-- >>> pretty (deg (5 :: Polynomial 1 Integer))
-- 0
-- >>> pretty (deg (0 :: Polynomial 1 Integer))
-- -∞
-- >>> pretty (deg (var 3 * var 5 + 4 :: Polynomial 10 Integer))
-- 2
deg :: Polynomial n a -> Degree
deg (Polynomial p) | M.null p = neginf
                   | otherwise = findeg . maximum . map degM . M.keys $ p

pick :: (a -> b -> Ordering) -> (a -> c) -> (b -> c) -> (a -> b -> c) ->
  a -> b -> c
pick cmp fromA fromB fromAB x y = case cmp x y of
  LT -> fromB y
  EQ -> fromAB x y
  GT -> fromA x

maximals :: Foldable t => (a -> a -> Ordering) -> t a -> [a]
maximals cmp = foldl' (flip step) []
  where step = pick (\a as -> maybe GT (cmp a) (listToMaybe as)) pure id (:)

-- | A list of the maximal-degree, or /leading/, 'Term's of a 'Polynomial'.
--
-- >>> p = (var 0 + 5 * var 1 + 2) * (var 0 + 3) :: Polynomial 2 Integer
-- >>> pretty p
-- x₀² + 5 x₀ x₁ + 5 x₀ + 15 x₁ + 6
-- >>> pretty (leading p)
-- [x₀², 5 x₀ x₁]
leading :: Polynomial n a -> [Term n a]
leading p = maximals (comparing (degM . getM)) (terms p)

-- | Check whether a 'Polynomial' is homogeneous—i.e., whether all of its
-- 'Monomial's have the same degree.
--
-- >>> homogeneous (var 1 + var 2 :: Polynomial 3 Integer)
-- True
-- >>> homogeneous (var 1 + var 2 + 5 :: Polynomial 3 Integer)
-- False
homogeneous :: Polynomial n a -> Bool
homogeneous = allSame . map degM . M.keys . _getCoeffs
  where allSame [] = True
        allSame (x:xs) = all (== x) xs

-- | Convert a 'Term' in @n@ variables to a 'Term' in @n + 1@ variables that
-- just happens to have an exponent of 0 for xₙ.
widen1T :: KnownNat n => Term n a -> Term (n + 1) a
widen1T (T (m, coeff)) = T (widen1M m, coeff)

-- | Convert a 'Polynomial' in @n@ variables to a 'Polynomial' in @n + 1@
-- variables that just happens to have an exponent of 0 for xₙ in each 'Term'.
widen1 :: KnownNat n => Polynomial n a -> Polynomial (n + 1) a
widen1 (Polynomial p) = Polynomial (M.mapKeys widen1M p)


-- | Free magmas.
--
-- A @'Grouping' a@ is a binary tree with @a@s at the leaves. Such a tree is
-- meant to represent a syntax tree for an expression built from constants and
-- a single binary operator. This lets us carry around expressions as values
-- before choosing to perform the actual expansion, which is important when the
-- result of the expansion is very large. Since our primary use cases are
-- operations that are associative, we could alternatively just use lists of
-- factors for this purpose, but it turns out that choices of associativity
-- actually make major /performance/ differences, even if they have no impact
-- on the resulting value.
--
-- 'Grouping' has a 'Foldable' instance, so 'foldMap' works on it, as well as
-- many functions like 'product' and 'maximum'. But see also 'foldG' if you
-- need to interpret 'Prod' as an operation that doesn't form a 'Monoid'.
--
-- >>> concat ((Leaf "hello " `Prod` Leaf "world") `Prod` Leaf "!")
-- "hello world!"
data Grouping a = Leaf a | Prod (Grouping a) (Grouping a)
  deriving (Eq, Ord, Show, Read, Functor, Foldable, Traversable)
infix `Prod`

instance Applicative Grouping where pure = return; (<*>) = ap
-- | '>>=' for 'Grouping' lets us map over leaves, replacing them with new
-- sub-'Grouping's. Alternatively: a @'Grouping' ('Grouping' a)@ is effectively
-- a @'Grouping' a@ with some of the subtrees marked, and 'join' erases this
-- marking.
instance Monad Grouping where
  return = Leaf
  Leaf a >>= k = k a
  Prod x y >>= k = Prod (x >>= k) (y >>= k)

ppGrouping :: (a -> Doc ann) -> Grouping a -> Doc ann
ppGrouping ppItem = go
  where p = nest 1 . parens
        go (Leaf a) = ppItem a
        go (Prod x y) = p (go x) <> " *" <> line <> p (go y)
-- | When pretty-printing, we visualize the 'Prod' constructor as a @*@ and
-- parenthesize each subexpression, including leaves.
--
-- >>> pretty ((Leaf "x" `Prod` Leaf "y") `Prod` Leaf "z")
-- ((x) *
--  (y)) *
-- (z)
instance Pretty a => Pretty (Grouping a) where pretty = ppGrouping pretty

-- | Fold over a 'Grouping'—i.e., evaluate the term it represents using some
-- interpretation for 'Prod'. This should be the same as 'Data.Foldable.fold',
-- but with an explicit argument instead of a 'Monoid' constraint, no need for
-- a 'mempty', and no obligation that the operation be associative.
--
-- >>> g = (Leaf 1 `Prod` Leaf 2) `Prod` Leaf 3 :: Grouping Int
-- >>> pretty g
-- ((1) *
--  (2)) *
-- (3)
-- >>> product g == foldG (*) g
-- True
-- >>> foldG (-) g
-- -4
foldG :: (a -> a -> a) -> Grouping a -> a
foldG _ (Leaf a) = a
foldG p (Prod x y) = foldG p x `p` foldG p y

-- | @'foldMapG' l p@ is a fused version of @'foldG' p . 'fmap' l@.
foldMapG :: (a -> b) -> (b -> b -> b) -> Grouping a -> b
foldMapG l _ (Leaf a) = l a
foldMapG l p (Prod x y) = foldMapG l p x `p` foldMapG l p y

-- | Strict 'foldG'. Applies the supplied operation /strictly/ (using '$!').
--
-- >>> u = Leaf "hello" `Prod` Leaf (error "bang!")
-- >>> foldG const u
-- "hello"
-- >>> foldG' const u
-- "*** Exception: bang!
-- CallStack (from HasCallStack):
--   error, called at <interactive>:... in interactive:...
foldG' :: (a -> a -> a) -> Grouping a -> a
foldG' _ (Leaf a) = a
foldG' p (Prod x y) = (p $! foldG' p x) $! foldG' p y

-- | @'foldMapG'' l p@ is a fused version of @'foldG'' p . 'fmap' (l '$!')@.
foldMapG' :: (a -> b) -> (b -> b -> b) -> Grouping a -> b
foldMapG' l _ (Leaf a) = l $! a
foldMapG' l p (Prod x y) = (p $! foldMapG' l p x) $! foldMapG' l p y

-- | @'groupR' x xs@ puts the elements of @xs@ into the leaves of a
-- right-associated 'Grouping', with @x@ at the far right.
--
-- >>> pretty (groupR 'x' "abcd")
-- (a) *
-- ((b) *
--  ((c) *
--   ((d) *
--    (x))))
-- >>> pretty (groupR 'x' "")
-- x
groupR :: Foldable t => a -> t a -> Grouping a
groupR z = foldr (\x g -> Prod (Leaf x) g) (Leaf z)
-- | @'groupL' x xs@ puts the elements of @xs@ into the leaves of a
-- left-associated 'Grouping', with @x@ at the far left.
--
-- >>> pretty (groupL 'x' "abc")
-- (((x) *
--   (a)) *
--  (b)) *
-- (c)
-- >>> pretty (groupL 'x' "")
-- x
groupL :: Foldable t => a -> t a -> Grouping a
groupL z = foldl (\g x -> Prod g (Leaf x)) (Leaf z)
-- | @'groupL' x xs@ puts the elements of @xs@ into the leaves of a
-- balanced 'Grouping', or produces @'Leaf' x@ if @xs@ is empty.
--
-- >>> pretty (groupB 'x' "abcde")
-- ((a) *
--  (b)) *
-- ((c) *
--  ((d) *
--   (e)))
-- >>> pretty (groupB 'x' "")
-- x
groupB :: Foldable t => a -> t a -> Grouping a
groupB z = go . toList
  where go []  = Leaf z
        go [x] = Leaf x
        go xs  = Prod (go xs1) (go xs2)
          where (xs1, xs2) = splitAt (length xs `div` 2) xs

-- | @'divide' d g@ removes each element of @d@ from @g@ and replaces it with a
-- 'Nothing', /taking multiplicity into account/. The leftmost occurrences of
-- an element in @g@ are removed first. An overall result of 'Nothing'
-- indicates that @g@ did not contain all of the elements to be removed.
--
-- >>> extendB s n = groupB 'x' (take n (cycle s))
-- >>> pretty (extendB "abc" 10)
-- (((a) *
--   (b)) *
--  ((c) *
--   ((a) *
--    (b)))) *
-- (((c) *
--   (a)) *
--  ((b) *
--   ((c) *
--    (a))))
-- >>> pretty (divide (extendB "ab" 4) (extendB "abc" 10))
-- ((() *
--   ()) *
--  ((c) *
--   (() *
--    ()))) *
-- (((c) *
--   (a)) *
--  ((b) *
--   ((c) *
--    (a))))
-- >>> pretty (divide "aaaaa" (extendB "abc" 10))
-- <BLANKLINE>
divide :: (Foldable t, Ord a) =>
  t a -> Grouping a -> Maybe (Grouping (Maybe a))
divide d0 g = if MS.null d1 then Just g' else Nothing
  where (g', d1) = runState (traverse replace1 g) (MS.fromList (toList d0))
        replace1 a = do
          mem <- gets (MS.member a)
          if mem then do modify (MS.delete a); return Nothing
                 else return (Just a)

-- | @'divideReplace' z d g@ is like @'divide' d g@, but instead of putting
-- 'Nothing's in place of the removed elements, they are replaced with @z@s.
--
-- >>> g = groupR 1 [2, 2, 5] :: Grouping Integer
-- >>> pretty g
-- (2) *
-- ((2) *
--  ((5) *
--   (1)))
-- >>> pretty (divideReplace 1 [2, 5] g)
-- (1) *
-- ((2) *
--  ((1) *
--   (1)))
-- >>> (product g, product <$> (divideReplace 1 [2, 5] g))
-- (20,Just 2)
divideReplace :: (Foldable t, Ord a) =>
  a -> t a -> Grouping a -> Maybe (Grouping a)
divideReplace z d0 g = fmap (fromMaybe z) <$> divide d0 g


-- | We think of an @'Expr' n a@ as an /expression/ for a polynomial in @n@
-- variables with coefficients from @a@, with multiplication deferred.
type Expr n a = Grouping (Polynomial n a)

-- | Check whether the 'Polynomial' given by an 'Expr' is homogeneous, but
-- /without/ actually multiplying it out. We should have
-- @'homogeneousE' e == 'homogeneous' ('product' e)@, but the former is very
-- cheap, while the latter is potentially extremely expensive.
--
-- >>> homogeneousE (Leaf (var 3 + var 0) `Prod` Leaf 5 :: Expr 6 Integer)
-- True
-- >>> homogeneousE (Leaf (var 3 + 2) `Prod` Leaf (var 1) :: Expr 6 Integer)
-- False
homogeneousE :: Expr n a -> Bool
homogeneousE = all homogeneous

-- | The degree of the 'Polynomial' given by an 'Expr', /without/ actually
-- multiplying it out. We should have
-- @'degE' e == 'deg' ('product' e)@, but the former is very cheap, while the
-- latter is potentially extremely expensive.
--
-- >>> pretty (degE (Leaf (var 0 + 1) `Prod` Leaf (var 0 - 1) :: Expr 1 Integer))
-- 2
degE :: Expr n a -> Degree
degE = foldMap deg

-- | Convert an 'Expr' in @n@ variables to an 'Expr' in @n + 1@ variables that
-- just happens to have an exponent of 0 for xₙ in each 'Term'.
widen1E :: KnownNat n => Expr n a -> Expr (n + 1) a
widen1E = fmap widen1


{- $bounded
   Say we have a polynomial p over a ring R, in n variables, and sets A_i ⊆ R
   of cardinalities k_i = |A_i|, for 0 ≤ i < n. We would like to apply the
   non-vanishing corollary to conclude that p does not vanish on all of
   ∏_i A_i. Then to know whether this is possible, it suffices to know p's
   equivalence class modulo the ideal I = (x₀^k₀, ..., xₙ₋₁^kₙ₋₁) and p's
   degree. This is because any two polynomials congruent mod I have the same
   set of monomials whose exponents are strictly bounded by the k_is, and once
   we know which such monomials p has, we can use our knowledge of its degree
   to check whether any are leading.

   Alternatively, if we know that p is homogeneous, then all terms are leading,
   so it suffices for p to be nonzero mod I.

   This suggests a way of working with polynomials that we want to apply the
   non-vanishing corollary to, cheaper than computing all of their terms:
   compute with their minimal representatives mod I. The next few functions
   provide facilities for doing this.

   As long as each k_i > 0, we can equivalently describe the ideal I as
   consisting of the polynomials whose monomials all fail to divide a
   particular bounding monomial b = ∏_i x_i^(k_i - 1). Then the minimal
   representatives mod I are the polynomials whose monomials all /do/ divide b.
   Since the non-vanishing corollary is rather vacuous when one of the A_i is
   empty, and this monomial is a convenient representation, it is the notion
   used by these functions.

   Together with turning 'Prelude.*'s that define 'Polynomial's into 'Prod's
   that define 'Expr's, multiplication with least representatives—"bounded
   multiplication"—forms the primary technique of this project. In the
   "Math.Alspach" module, some functions with exponential complexity have had
   the base of their complexity dramatically reduced by replacing a naïve
   multiplication with a 'boundedProductE' and a good choice of 'Expr'. This
   can make the difference between 100s and 0.33s on an input as small as n=7,
   and the ratio only gets bigger!

   TODO: The thing about bases /seems/ to be roughly true, but it's a pretty
   bold claim, and I should *really* work it through. Actually, I think some
   complexities are /super/exponential!
-}

-- | @'reduce' b p@ is the minimal representative of @p@ modulo the ideal
-- associated to the bounding 'Monomial' @b@. Equivalently, it is @p@ with all
-- of the terms removed whose 'Monomial' does not divide @b@.
--
-- >>> p = 3 * var 0 * var 1 + 5 * var 0 * var 0 + 2 * var 0 + 4 * var 1 + 1 :: Polynomial 2 Integer
-- >>> pretty (reduce (varM 0) p)
-- 2 x₀ + 1
-- >>> pretty (reduce (varM 1) p)
-- 4 x₁ + 1
-- >>> pretty (reduce (varM 0 <> varM 1) p)
-- 3 x₀ x₁ + 2 x₀ + 4 x₁ + 1
reduce :: KnownNat n => Monomial n -> Polynomial n a -> Polynomial n a
reduce bound = filterTerms (\(T (m, _)) -> m <<= bound)

-- | @'boundedMul' b p q@ is a fused version of @'reduce' b (p * q)@.
--
-- >>> pretty (boundedMul (varM 0) (var 0 + 1) (var 0 + 1) :: Polynomial 1 Integer)
-- 2 x₀ + 1
boundedMul :: (KnownNat n, Num a, Eq a) =>
  Monomial n -> Polynomial n a -> Polynomial n a -> Polynomial n a
boundedMul bound p q = fromTerms $
  [T (mp <> mq, coeffp * coeffq) |
   T (mp, coeffp) <- terms p, T (mq, coeffq) <- terms q, (mp <> mq) <<= bound]

-- | Multiply out an 'Expr', reduced mod an ideal given by a bounding monomial.
-- We should have @'boundedProductE' b e == 'reduce' ('product' e)@. However,
-- the former may be much, much cheaper, because it takes least representatives
-- at each subterm instead of only at the end, possibly cutting down on the
-- exploding number of terms.
--
-- >>> e = replicate 100 (var 0 + var 1 + var 2 :: Polynomial 3 Integer)
-- >>> b = powM (varM 0) 100 <> varM 1 <> varM 2
-- >>> pretty (boundedProductE b (groupB 1 e))
-- x₀¹⁰⁰ + 100 x₀⁹⁹ x₁ + 100 x₀⁹⁹ x₂ + 9900 x₀⁹⁸ x₁ x₂
boundedProductE :: (KnownNat n, Num a, Eq a) =>
  Monomial n -> Expr n a -> Polynomial n a
boundedProductE bound = foldG (boundedMul bound)

-- | Do a 'boundedProductE' on an arbitrary 'Foldable' collection of
-- 'Polynomial's, defaulting to right associativity. You may get much better
-- performance with other ways of associating the product; if performance is
-- important, try experimenting with explicitly grouping the collection into an
-- 'Expr' yourself and then using 'boundedProductE'.
boundedProduct :: (KnownNat n, Num a, Eq a, Foldable t) =>
  Monomial n -> t (Polynomial n a) -> Polynomial n a
boundedProduct bound = boundedProductE bound . groupR 1


{- $multisets
   The remainder of this module implements some tools for doing bounded
   multiplication by keeping a multiset of factors and selecting ones to
   multiply next. It doesn't seem to work quite as well as a good choice of
   'Expr'.

   TODO: Document this stuff.
-}

-- This newtype is just so that we can write a new Ord instance.
newtype Factor n a = Factor {getFactor :: Polynomial n a}
  deriving (Eq, Show, Read, Functor, Foldable, Traversable, Num, Pretty)

lexico :: Ord a => [a -> a -> Ordering] -> a -> a -> Ordering
lexico [] x y = compare x y
lexico (cmp:cmps) x y = case cmp x y of
  EQ -> lexico cmps x y
  r -> r

instance Ord a => Ord (Factor n a) where
  compare = lexico [comparing length] `on` getFactor

type FactorMS n a = MultiSet (Factor n a)
-- A 'SelFun' is a function that takes a multiset of polynomials and exercises
-- some judgment as to which should be multiplied together next. The
-- 'Selection' type exists purely to be its codomain; a 'SelFun' can result in
-- 'Empty' if it is given an empty multiset, 'Singleton p' if it is given a
-- singleton multiset containing 'p', and 'DoMul p q f' if it has decided that
-- 'p' and 'q' should be multiplied next, with 'f' being the remainder of the
-- multiset after they have been removed.
data Selection n a
  = Empty | Singleton (Polynomial n a)
  | DoMul (Polynomial n a) (Polynomial n a) (FactorMS n a)
  deriving (Show)
type SelFun n a = FactorMS n a -> Selection n a

instance (Pretty a, Num a, Eq a) => Pretty (Selection n a) where
  pretty Empty = "" -- TODO figure out precedence properly
  pretty (Singleton p) = "Singleton " <+> parens (pretty p)
  pretty (DoMul p q _) = "DoMul " <+> parens (pretty p) <+>
    parens (pretty q) <+> "[more factors]"

boundedProductMS :: (KnownNat n, Num a, Ord a) =>
  SelFun n a -> Monomial n -> MultiSet (Polynomial n a) -> Polynomial n a
boundedProductMS select bound = go . MS.map Factor
  where go fs = case select fs of
          Empty -> 1
          Singleton p -> p
          DoMul p q fs' ->
            let f = Factor (boundedMul bound p q)
            in go (MS.insert f fs')

boundedProductViaMS :: (KnownNat n, Num a, Ord a) =>
  SelFun n a -> Monomial n -> [Polynomial n a] -> Polynomial n a
boundedProductViaMS select bound = boundedProductMS select bound . MS.fromList

-- This seems to work so-so.
minmax :: (KnownNat n, Ord a) => SelFun n a
minmax fs = case MS.minView fs of
  Nothing -> Empty
  Just (Factor small, fs') -> case MS.maxView fs' of
    Nothing -> Singleton small
    Just (Factor large, fs'') -> DoMul small large fs''

