From Coq Require Import ssreflect ssrfun ssrbool.

Require Import mathcomp.ssreflect.tuple.
Require Import mathcomp.ssreflect.ssrnat.
Require Import mathcomp.ssreflect.fintype.
Require Import mathcomp.ssreflect.eqtype.
Require Import mathcomp.ssreflect.choice.
From mathcomp Require Import finfun.
From mathcomp Require Import finmap.
Require Import mathcomp.algebra.zmodp.
Require Import mathcomp.algebra.ssralg.
Require Import mathcomp.algebra.ssrint.
Require Import SsrMultinomials.mpoly.
Require Import FMaps.
Require Import FMapAVL.
Require Import mathcomp.ssreflect.seq.
Require Import mathcomp.ssreflect.bigop.

Set Bullet Behavior "Strict Subproofs".

Fixpoint tuple' (T : Type) (n : nat) : Type :=
  match n with
  | O => unit
  | S n' => T * tuple' T n'
  end.

Lemma I0_empty : forall {P} (i : 'I_0), P i.
Proof. move=> P [] //. Qed.

Fixpoint tnth' {T} {n} : tuple' T n -> 'I_n -> T :=
  match n with
  | O => fun _ o => I0_empty o
  | S n' => fun x o => match o with
                   | Ordinal O _ => x.1
                   | Ordinal (S m) L => tnth' x.2 (Ordinal (etrans (esym (ltnS _ _)) L))
                   end
  end.

Lemma tnth'_eq : forall {T} {n} (x y : tuple' T n), tnth' x =1 tnth' y -> x = y.
Proof.
  move=> T.
  elim; first by move=> [] [] //.
  move=> n IH [x1 x2] [y1 y2] E; f_equal.
  - move: (E ord0) => //.
  - apply: IH => i.
    move: (E (lift ord0 i)) => /=.
    rewrite [Ordinal _](ord_inj _ : _ = i) //.
Qed.

Fixpoint map_tuple' {T S} {n} (f : T -> S) : tuple' T n -> tuple' S n :=
  match n with
  | O => fun _ => tt
  | S n' => fun x => (f x.1, map_tuple' f x.2)
  end.

Lemma map_tuple'_spec : forall {T S} {n} {f : T -> S} {x : tuple' T n} {i : 'I_n},
    tnth' (map_tuple' f x) i = f (tnth' x i).
Proof.
  move=> T S.
  elim; first by move=> ? ?; apply: I0_empty.
  move=> n IH f x [[| m] L] //=.
Qed.

Fixpoint map2_tuple' {T S R} {n} (f : T -> S -> R)
  : tuple' T n -> tuple' S n -> tuple' R n :=
  match n with
  | O => fun _ _ => tt
  | S n' => fun x y => (f x.1 y.1, map2_tuple' f x.2 y.2)
  end.

Lemma map2_tuple'_spec : forall {T S R} {n} {f : T -> S -> R}
                           {x : tuple' T n} {y : tuple' S n} {i : 'I_n},
    tnth' (map2_tuple' f x y) i = f (tnth' x i) (tnth' y i).
Proof.
  move=> T S R.
  elim; first by move=> ? ? ?; apply: I0_empty.
  move=> n IH f x y [[| m] L] //=.
Qed.

Fixpoint foldr_tuple' {T S} {n} (f : T -> S -> S) (z : S) : tuple' T n -> S :=
  match n with
  | O => fun _ => z
  | S n' => fun x => f x.1 (foldr_tuple' f z x.2)
  end.

Lemma foldr_tuple'_big : forall {T} {n} {f : T -> T -> T} {z : T} {x : tuple' T n},
    foldr_tuple' f z x = \big[f/z]_i tnth' x i.
Proof.
  move=> T.
  elim; first by move=> *; rewrite big_ord0 //.
  move=> n IH f z x.
  rewrite [foldr_tuple' _ _ _]/= big_ord_recl IH.
  apply/f_equal/eq_bigr => /= i _.
    by apply/f_equal/ord_inj.
Qed.

Fixpoint iota' (n : nat) : tuple' 'I_n n :=
  match n with
  | O => tt
  | S n' => (ord0, map_tuple' (lift ord0) (iota' n'))
  end.
Definition tabulate {T} {n} (f : 'I_n -> T) : tuple' T n := map_tuple' f (iota' n).

Lemma iota'_spec : forall {n}, tnth' (iota' n) =1 id.
Proof.
  elim; first by apply: I0_empty.
  move=> n IH [[| m] L]; apply: ord_inj => //=.
  rewrite map_tuple'_spec IH //.
Qed.
Lemma tabulate_spec : forall {T n} {f : 'I_n -> T}, tnth' (tabulate f) =1 f.
Proof. move=> T n f i. rewrite /tabulate map_tuple'_spec iota'_spec //. Qed.

Definition t_from_mathcomp {T} {n} (t : n.-tuple T) : tuple' T n := tabulate (tnth t).
(*
  match n with
  | O => fun _ => tt
  | S n' => fun t => (thead t, t_from_mathcomp [tuple of behead t])
  end.
 *)
Definition t_to_mathcomp {T} {n} (x : tuple' T n) : n.-tuple T :=
  [tuple tnth' x i | i < n].
(*
  match n with
  | O => fun _ => [tuple]
  | S n' => fun x => [tuple of x.1 :: t_to_mathcomp x.2]
  end.
 *)
Lemma t_inverse1 : forall {T} {n} (t : n.-tuple T), t_to_mathcomp (t_from_mathcomp t) = t.
Proof. move=> T n t. apply/eq_from_tnth/(ftrans (tnth_mktuple _))/tabulate_spec. Qed.
Lemma t_inverse2 : forall {T} {n} (x : tuple' T n), t_from_mathcomp (t_to_mathcomp x) = x.
Proof. move=> T n x. apply/tnth'_eq/(ftrans tabulate_spec)/tnth_mktuple. Qed.

Fixpoint mathcomp_iota (n : nat) : n.-tuple 'I_n :=
  match n with
  | O => [tuple]
  | S n' => [tuple of ord0 :: map_tuple (lift ord0) (mathcomp_iota n')]
  end.

Fixpoint tuple'_eq {T} {n} (eq : relation T) : relation (tuple' T n) :=
  match n with
  | O => fun _ _ => True
  | S n' => fun x y => eq x.1 y.1 /\ tuple'_eq eq x.2 y.2
  end.
Fixpoint tuple'_lt {T} {n} (lt eq : relation T) : relation (tuple' T n) :=
  match n with
  | O => fun _ _ => False
  | S n' => fun x y => lt x.1 y.1 \/ (eq x.1 y.1 /\ tuple'_lt lt eq x.2 y.2)
  end.

Lemma tuple'_eq_eq : forall {T n} (x y : tuple' T n), tuple'_eq eq x y <-> x = y.
Proof.
  move=> T.
  elim; first by move=> [] [].
  move=> n /= IH [x1 x2] [y1 y2] /=.
  rewrite IH; split=> -[-> ->] //.
Qed.

Fixpoint inord' {m} (n : nat) : 'I_m.+1 :=
  match n, m with
  | O, _ | _, O => ord0
  | S n', S m' => lift ord0 (inord' n')
  end.
Definition seq_iota' (n : nat) : seq 'I_n := foldr_tuple' cons nil (iota' n).
Definition irange {c} (i j : 'I_c) : seq 'I_c :=
  [seq k <- seq_iota' c | i <= nat_of_ord k <= j].

Definition tuple'_compare {T} {n} (lt eq : relation T)
           (eq_sym : forall x y : T, eq x y -> eq y x)
         (compare : forall x y : T, Compare lt eq x y) :
  forall x y : tuple' T n, Compare (tuple'_lt lt eq) (tuple'_eq eq) x y.
Proof with auto.
  elim: n => /=.
  - move=> [] []; by apply: EQ.
  - move=> n IH x y.
    case: (compare x.1 y.1) => [L1 | E1 | G1].
    + apply: LT...
    + case: (IH x.2 y.2) => [L2 | E2 | G2].
      * apply: LT...
      * apply: EQ...
      * apply: GT...
    + apply: GT...
Defined.

Section tuple'_eqType.
  Variable T : eqType.

  Definition tuple'_eqMixin n : Equality.mixin_of (tuple' T n).
  Proof.
    elim: n => /=; first by exact [eqMixin of unit].
    move=> n Mix; apply: prod_eqMixin (EqType _ Mix).
  Defined.

  Canonical Structure tuple'_eqType n :=
    Eval hnf in EqType (tuple' T n) (tuple'_eqMixin n).
End tuple'_eqType.

Section tuple'_choiceType.
  Variable T : choiceType.

  Definition tuple'_choiceMixin n : Choice.mixin_of (tuple' T n).
  Proof.
    elim: n => /=; first by exact unit_choiceMixin.
    move=> n Mix; apply: prod_choiceMixin (ChoiceType _ Mix).
  Defined.

  Canonical Structure tuple'_choiceType n :=
    Eval hnf in ChoiceType (tuple' T n) (tuple'_choiceMixin n).
End tuple'_choiceType.

Module Type LexArgs.
  Declare Module O : OrderedType.
  Parameter n : nat.
End LexArgs.
Module MiniLex (A : LexArgs) <: MiniOrderedType.
  Include A.
  Module F := OrderedTypeFacts O.

  Definition t : Type := tuple' O.t n.
  Definition eq : t -> t -> Prop := tuple'_eq O.eq.
  Hint Unfold eq.
  Definition lt : t -> t -> Prop := tuple'_lt O.lt O.eq.
  Hint Unfold lt.

  Theorem eq_refl : forall x : t, eq x x.
  Proof. rewrite /t /eq. elim: n; firstorder using O.eq_refl. Qed.
  Theorem eq_sym : forall x y : t, eq x y -> eq y x.
  Proof. rewrite /t /eq. elim: n; firstorder using O.eq_sym. Qed.
  Theorem eq_trans : forall x y z : t, eq x y -> eq y z -> eq x z.
  Proof. rewrite /t /eq. elim: n; firstorder using O.eq_trans. Qed.
  Theorem lt_trans : forall x y z : t, lt x y -> lt y z -> lt x z.
  Proof.
    rewrite /t /lt. elim: n => //= n' IH.
    move=> x y z [L1 | [<- L1]] [L2 | [<- L2]];
            eauto using O.lt_trans.
  Qed.
  Theorem lt_not_eq : forall x y : t, lt x y -> ~ eq x y.
  Proof. rewrite /t /lt /eq. elim: n; firstorder using O.lt_not_eq. Qed.
  Definition compare : forall x y : t, Compare lt eq x y :=
    tuple'_compare O.lt O.eq O.eq_sym O.compare.
End MiniLex.
Module Lex (A : LexArgs) <: OrderedType.
  Module Mini := MiniLex A.
  Include MOT_to_OT Mini.
End Lex.

Module Type MapPolyArgs.
  Parameter n : nat.
  Parameter Ri : idomainType.
End MapPolyArgs.
Module MapPoly (A : MapPolyArgs).
  Include A.

  Definition monomial : Type := tuple' nat n.
  Module MonomialO <: OrderedType.
    Module Args <: LexArgs.
      Module O := Nat_as_OT.
      Definition n := n.
    End Args.
    Include Lex Args.
  End MonomialO.
  Module M := FMapAVL.Make MonomialO.
  Module MProps := Properties M.
  (* Module M := M0.Raw. *)
  Program Definition map_option {elt elt'}
          (f : M.key -> elt -> option elt') (m : M.t elt) : M.t elt' :=
    @M.Bst _ (M.Raw.map_option f (M.this m)) _.
  Next Obligation.
    apply: M.Raw.Proofs.map_option_bst; last by apply: M.is_bst.
    move=> x x' d /tuple'_eq_eq -> //.
  Qed.

  Lemma eqke_eq : forall {T : eqType},
      pointwise_relation _ (pointwise_relation _ iff)
                         (@M.eq_key_elt T) (@eq (M.key * T)).
  Proof.
    move=> T [t1 t2] [t'1 t'2].
    split=> [[/= E ->] | -> //]; f_equal.
      by apply/tuple'_eq_eq.
  Qed.
  Lemma mathcomp_in1 : forall {T : eqType} {R : relation T} `{!Reflexive R}
                         {t : T} {l : seq T},
      t \in l -> InA R t l.
  Proof.
    move=> T R ? t; elim=> // t' l IH.
    rewrite in_cons => /orP [/eqP <- | /IH IH']; constructor; done.
  Qed.
  Lemma mathcomp_in : forall {T : eqType} {t : M.key * T} {l : seq (M.key * T)},
      InA (@M.eq_key_elt T) t l <-> t \in l.
  Proof.
    move=> T t.
    elim=> [| t' l IH]; first by rewrite in_nil InA_nil //.
    rewrite in_cons InA_cons IH -(rwP orP) -(rwP eqP) eqke_eq //.
  Qed.
  Lemma mathcomp_uniq1 : forall {T : eqType} {R : relation T} `{!Reflexive R} {l : seq T},
      NoDupA R l -> uniq l.
  Proof.
    move=> T R ? l; elim: l / => //= a l NotIn _ ?.
    apply/andP; split; last by done.
    apply/negP => /mathcomp_in1 //.
  Qed.
  Lemma mathcomp_uniq : forall {T : eqType} {l : seq (M.key * T)},
      uniq l <-> NoDupA (@M.eq_key_elt T) l.
  Proof.
    move=> T l; split; last by apply: mathcomp_uniq1.
    elim: l => //= a l IH /andP [/negP /mathcomp_in NotIn /IH Uniq].
      by constructor.
  Qed.

  Definition polynomial : Type := M.t Ri.
  Definition term : Type := monomial * Ri.

  Definition ok (p : polynomial) : Prop := forall m coeff, M.MapsTo m coeff p -> coeff != 0%R.
  Definition poly_eq (p q : polynomial) : Prop := M.Equal p q /\ ok p.

  Lemma poly_eq_dom : forall p, poly_eq p p <-> ok p.
  Proof. firstorder. Qed.
  Lemma poly_eq_dom' : forall p q, poly_eq p q -> ok p.
  Proof. firstorder. Qed.
  Instance ok_proper : Proper (poly_eq ==> iff) ok.
  Proof.
    move=> p p' [/MProps.F.Equal_mapsto_iff E Ok].
    split=> // _ m coeff.
    rewrite -E.
    apply: Ok.
  Qed.

  Instance poly_eq_per : PER poly_eq.
  Proof.
    constructor.
    - move=> p q [E Ok].
      split; first by symmetry.
      move=> m coeff.
      rewrite -(proj1 (MProps.F.Equal_mapsto_iff _ _) E).
      apply: Ok.
    - move=> p q r [E1 Ok1] [E2 Ok2].
      split.
      + by etransitivity.
      + apply: poly_eq_dom' (conj E1 Ok1).
  Qed.

  Open Scope fsfun_scope.
  Open Scope fset.

  Definition avl_dom {A} (m : M.t A) : {fset monomial}
    := M.fold (fun k v f => k |` f) m fset0.
  Lemma avl_dom_spec {A} (m : M.t A) (k : monomial)
    : reflect (M.In k m) (k \in avl_dom m).
  Proof.
    apply/iff_reflect; rewrite -/(is_true _) /avl_dom; elim/MProps.fold_rec: _.
    - move=> m' /MProps.elements_Empty E_el.
      rewrite MProps.F.in_find_iff MProps.F.elements_o E_el //=.
    - move=> k' v /= S m' m'' ? ? Add IH.
      rewrite MProps.F.in_find_iff Add.
      case: (eqP (x:=k) (y:=k')) => [E | NE].
      + subst k'; rewrite fset1U1 MProps.F.add_eq_o //.
      + rewrite -(rwP (fset1UP _ _ _)) -IH MProps.F.add_neq_o.
        * rewrite MProps.F.in_find_iff; tauto.
        * by apply/tuple'_eq_eq/nesym.
  Qed.

  Definition avl_from_mathcomp {A : choiceType} (a : A)
             (f : {fsfun monomial -> A with a}) : M.t A :=
    MProps.of_list [seq (val k, f (val k)) | k : finsupp f].
  Definition avl_to_mathcomp {A : choiceType} (a : A)
             (m : M.t A) : {fsfun monomial -> A with a} :=
    [fsfun k in avl_dom m => odflt a (M.find k m)].

  Lemma a_inverse1 : forall {A : choiceType} (a : A),
      cancel (avl_from_mathcomp a) (avl_to_mathcomp a).
  Proof.
    rewrite /avl_to_mathcomp.
    move=> A a /= f; apply/fsfunP => /= k.
    rewrite fsfun_ffun; case: insubP => /= [u In <- |].
    - give_up.
    - move=> NotIn; rewrite fsfun_dflt //.
      move: NotIn; apply/contra => In; apply/avl_dom_spec.
      exists (f k).
      rewrite -/(M.MapsTo k _ _) /avl_from_mathcomp MProps.of_list_1.
      + apply/mathcomp_in/mapP; exists [` In] => //=; apply/mem_enum.
      + have /mathcomp_uniq : uniq [seq (val k, f (val k)) | k : finsupp f]. {
          rewrite map_inj_uniq; first by apply: enum_uniq.
          move=> k1 k2 [/val_inj //].
        }
        unfold image_mem; elim: (enum _) => //= k' l IH.
        inversion_clear 1 as [| ? ? NotIn H]; constructor; last by auto.
        contradict NotIn; elim: l NotIn {IH H} => /=.
  Qed.
  Lemma a_inverse2 : forall {A} {m : M.t A}, M.Equal (avl_from_mathcomp (avl_to_mathcomp m)) m.
  Proof.
    rewrite /avl_from_mathcomp /avl_to_mathcomp.
    move=> A m0; elim/MProps.fold_rec: _.
    - move=> m Emp.
      rewrite /= [[seq _ | _ : _]]size0nil; last by rewrite size_image /= cardfT0 //.
      rewrite /= -[m]MProps.fold_identity MProps.fold_Empty //.
    - move=> k v f m m' MapsTo NotIn Add E.
      change (M.Equal m' (M.add k v m)) in Add.
      rewrite Add -E /=.
  Abort.

  (*

  Instance ina_proper {A} :
    Proper (pointwise_relation A (pointwise_relation A iff) ==> eq ==> eq ==> iff) (@InA A).
  Proof.
    move=> R R' E a _ <- l _ <-.
    elim: l a; first by split; inversion 1.
    move=> a a' l IH.
    split; inversion 1; firstorder.
  Qed.
  Instance nodupa_proper {A} :
    Proper (pointwise_relation A (pointwise_relation A iff) ==> eq ==> iff) (@NoDupA A).
  Proof.
    move=> R R' E l _ <-.
    elim: l; first by auto.
    move=> a l IH.
    split; inversion 1; constructor.
    - rewrite -E //.
    - intuition.
    - rewrite E //.
    - intuition.
  Qed.
   *)

  Definition m_from_mathcomp (m : 'X_{1..n}) : monomial := t_from_mathcomp m.
  Definition m_to_mathcomp (m : monomial) : 'X_{1..n} := [multinom t_to_mathcomp m].

  Lemma m_inverse1 : cancel m_from_mathcomp m_to_mathcomp.
  Proof. move=> [exps]. rewrite /m_to_mathcomp /m_from_mathcomp t_inverse1 //. Qed.
  Lemma m_inverse2 : cancel m_to_mathcomp m_from_mathcomp.
  Proof. move=> m. rewrite /m_from_mathcomp /m_to_mathcomp mnm_valK t_inverse2 //. Qed.
  Lemma m_injective1 : injective m_from_mathcomp.
  Proof. apply/can_inj/m_inverse1. Qed.
  Lemma m_injective2 : injective m_to_mathcomp.
  Proof. apply/can_inj/m_inverse2. Qed.
  Hint Resolve m_inverse1 m_inverse2 m_injective1 m_injective2.

  Definition m_mul : monomial -> monomial -> monomial := map2_tuple' (fun n m => n + m).
  Definition m_div : monomial -> monomial -> monomial := map2_tuple' (fun n m => n - m).
  Definition m_cmp (m1 m2 : monomial) : bool :=
    foldr_tuple' andb true (map2_tuple' leq m1 m2).

  Lemma m_mul_spec1 : {morph m_to_mathcomp : m1 m2 / m_mul m1 m2 >-> (m1 + m2)%MM}.
  Proof.
    move=> m1 m2.
    rewrite /m_to_mathcomp mnmP => i.
    rewrite mnmDE !multinomE !tnth_mktuple map2_tuple'_spec //.
  Qed.
  Lemma m_mul_spec2 : {morph m_from_mathcomp : m1 m2 / (m1 + m2)%MM >-> m_mul m1 m2}.
  Proof.
    move=> m1 m2.
    apply/tnth'_eq => i.
    rewrite map2_tuple'_spec !tabulate_spec -!mnm_tnth mnmDE //.
  Qed.

  Definition m_deg : monomial -> nat := foldr_tuple' (fun n m => n + m) 0.
  Definition var_m (i : 'I_n) : monomial := tabulate (fun j => if j == i then 1 else 0).
  Definition const_exp_m (m : nat) : monomial := tabulate (fun _ => m).
  Definition m_one : monomial := const_exp_m 0.
  Definition eval_m (m : monomial) (v : tuple' Ri n) : Ri :=
    foldr_tuple' *%R 1%R (map2_tuple' (fun x n => x^+n)%R v m).

  Lemma m_deg_spec : forall m, m_deg m = mdeg (m_to_mathcomp m).
  Proof.
    move=> m.
    rewrite /m_deg /m_to_mathcomp mdegE foldr_tuple'_big.
    apply: eq_bigr => i _.
    rewrite mnmE //.
  Qed.
  Lemma var_m_spec : forall i, m_to_mathcomp (var_m i) = U_(i)%MM.
  Proof.
    move=> i; apply/mnmP => j.
    rewrite /m_to_mathcomp mnmE tabulate_spec mnm1E eq_sym //.
  Qed.
  Lemma m_one_spec : m_to_mathcomp m_one = 0%MM.
  Proof. apply/mnmP => i. rewrite mnmE tabulate_spec mnm0E //. Qed.
  Lemma eval_m_spec : forall m v, 'X_[m_to_mathcomp m].@[tnth' v] = eval_m m v.
  Proof.
    move=> m v.
    rewrite mevalX /eval_m foldr_tuple'_big.
    apply: eq_bigr => i _.
    rewrite map2_tuple'_spec mnmE //.
  Qed.

  Definition zero : polynomial := M.empty Ri.
  Definition terms : polynomial -> seq term := @M.elements _.
  Definition m_supp (p : polynomial) : seq monomial := map fst (terms p).
  Definition coeff_for (m : monomial) (p : polynomial) : Ri := odflt 0%R (M.find m p).
  Definition from_terms : seq term -> polynomial :=
    let add1 '(m, coeff) p := if coeff != 0%R then M.add m coeff p else p in
    foldr add1 zero.
  Definition filter_terms (f : term -> bool) : polynomial -> polynomial :=
    map_option (fun m coeff => if f (m, coeff) then Some coeff else None).

  Definition p_from_mathcomp (p : {mpoly Ri[n]}) : polynomial :=
    MProps.of_list [seq (m_from_mathcomp m, p@_m)%R | m <- msupp p].
  Definition p_to_mathcomp (p : polynomial) : {mpoly Ri[n]} :=
    \sum_(t <- terms p) t.2 *: 'X_[m_to_mathcomp t.1].

  Lemma p_ok : forall {p}, ok (p_from_mathcomp p).
  Proof.
    move=> p m coeff MapsTo.
    suff H : NoDupA (@M.eq_key Ri) [seq (m_from_mathcomp m, (p@_m)%R) | m <- msupp p]. {
      move: MapsTo => /(MProps.of_list_1 _ _ H) /mathcomp_in /mapP [m'].
      rewrite mcoeff_msupp => NE [_ ->] //.
    }
    elim: (msupp p) (msupp_uniq p); first by constructor.
    move=> m' Supp IH /= /andP [NotIn Uniq]; constructor; last by apply: IH.
    elim: {IH Uniq} Supp NotIn; first by inversion 2.
    move=> m'' Supp IH; rewrite in_cons negb_or /= => /andP [/eqP NE NotIn].
    inversion 1 as [y l E | y l In']; subst.
    - move: E => /tuple'_eq_eq /m_injective1 //.
    - by apply: IH.
  Qed.
  Hint Resolve p_ok.

  Definition ok_t (t : term) : bool := t.2 != 0%R.
  Definition ok_ts : seq term -> bool := all ok_t.

  Lemma zero_spec : p_to_mathcomp zero = 0%R.
  Proof. rewrite /p_to_mathcomp /terms MProps.elements_empty big_nil //. Qed.

  Section P.
    Local Open Scope ring_scope.
    Variable p : polynomial.
    Hypothesis Ok : ok p.

    Lemma terms_uniq : uniq (terms p).
    Proof. apply/mathcomp_uniq/MProps.NoDupA_eqk_eqke/M.elements_3w. Qed.
    Hint Resolve terms_uniq.
    Lemma m_supp_uniq : uniq (m_supp p).
    Proof.
      apply: (@map_uniq _ _ (fun k => option_map (pair k) (M.find k p))).
      rewrite -map_comp (proj1 (eq_in_map _ Some _) _);
        first by rewrite map_inj_uniq //; apply: @Some_inj.
      move=> /= [m coeff] /mathcomp_in In /=.
      rewrite (M.find_1 (M.elements_2 In)) //.
    Qed.
    Hint Resolve m_supp_uniq.

    Lemma m_supp_from_terms {m coeff} : (m, coeff) \in terms p -> m \in m_supp p.
    Proof. apply: map_f. Qed.
    Hint Resolve m_supp_from_terms.

    Lemma in_terms {m coeff} : (m, coeff) \in terms p <-> M.MapsTo m coeff p.
    Proof. rewrite -mathcomp_in MProps.F.elements_mapsto_iff //. Qed.
    Lemma in_m_supp {m} : m \in m_supp p <-> M.MapsTo m (coeff_for m p) p.
    Proof.
      rewrite -(rwP mapP) /=.
      split.
      - move=> {m} [[m coeff] /mathcomp_in In /= ->].
        rewrite /coeff_for (M.find_1 (M.elements_2 In)) /=.
        apply/M.elements_2/In.
      - move=> MapsTo; exists (m, coeff_for m p); last by done.
        apply/mathcomp_in/M.elements_1/MapsTo.
    Qed.

    Lemma terms_from_supp : terms p = [seq (m, coeff_for m p) | m <- m_supp p].
    Proof.
      rewrite -map_comp map_id_in // => /= -[m coeff] /= /mathcomp_in In.
      rewrite /coeff_for (M.find_1 (M.elements_2 In)) //.
    Qed.

    Lemma terms_spec1 {m coeff} :
      (m, coeff) \in terms p <-> coeff != 0 /\
                               (p_to_mathcomp p)@_(m_to_mathcomp m) = coeff.
    Proof.
      rewrite /p_to_mathcomp GRing.raddf_sum {2}terms_from_supp big_map /=
              (eq_bigr (fun m' => if m' == m then coeff_for m' p else 0)). 2: {
        move=> m' _.
        rewrite mcoeffZ mcoeffX inj_eq //.
        case: eqP; rewrite ?GRing.mulr1 ?GRing.mulr0 //.
      }
      rewrite -big_mkcond -big_filter /= //.
      split.
      - move=> In; move: (In) => /in_terms MapsTo.
        rewrite filter_pred1_uniq; eauto.
        rewrite big_seq1 /coeff_for (M.find_1 MapsTo); eauto.
      - move=> [Nonzero].
        case E: (m \in m_supp p).
        + rewrite filter_pred1_uniq // big_seq1 /= => <-.
          apply/mathcomp_in/M.elements_1/in_m_supp/E.
        + rewrite (eq_in_filter (a2:=pred0)). 2: {
            move=> m' /= In.
            apply/negP => /eqP ?; subst.
            move: In E => -> //.
          }
          rewrite filter_pred0 big_nil => ?; subst.
          move: Nonzero => /negP //.
    Qed.

    Lemma m_supp_spec1 :
      perm_eq (map m_to_mathcomp (m_supp p)) (msupp (p_to_mathcomp p)).
    Proof.
      apply: uniq_perm; [by rewrite map_inj_uniq | done |].
      move=> /= m.
      rewrite -map_comp mcoeff_msupp; apply/mapP/idP.
      - move=> {m} [[m coeff] /= /terms_spec1 [Nonzero E] ->]; rewrite E //.
      - move=> Nonzero; eexists (m_from_mathcomp m, (p_to_mathcomp p)@_m); last by auto.
        apply/terms_spec1; split; first by done.
        rewrite m_inverse1 //.
    Qed.

    (*
    Lemma p_inverse2 : poly_eq (p_from_mathcomp (p_to_mathcomp p)) p.
    Proof.
      split; last by auto.
      move=> m.
      rewrite /p_from_mathcomp.
     *)
  End P.

  Lemma p_inverse1 : forall p, p_to_mathcomp (p_from_mathcomp p) = p.
  Proof.
    (*
    move=> p.
    rewrite [p_to_mathcomp _]mpolyE -(perm_big _ (m_supp_spec1 _ p_ok)) /= big_map.
    move: (perm_map m_from_mathcomp (m_supp_spec1 (p_from_mathcomp p) p_ok)).
    rewrite -map_comp map_id_in => [PE | m _]; last by auto.
    rewrite (perm_big _ PE) {PE} big_map /=.
    rewrite /p_from_mathcomp.
     *)


    move=> p.
    rewrite /p_to_mathcomp /p_from_mathcomp terms_from_supp big_map /=.
    set from := MProps.of_list _.
    rewrite (eq_big_seq (fun m' => p@_(m_to_mathcomp m') *:
                                 'X_[m_to_mathcomp m']))%R. 2: {
      have H : NoDupA (@M.eq_key Ri) [seq (m_from_mathcomp m, (p@_m)%R) | m <- msupp p] by give_up.
      move=> m' /in_m_supp /(MProps.of_list_1 _ _ H) /mathcomp_in /mapP [m'' _ [-> ->]].
      rewrite m_inverse1 //.
    }
    rewrite -(big_map m_to_mathcomp xpredT (fun m' => p@_m' *: 'X_[m'])%R)
               (perm_big _ (m_supp_spec1 _ p_ok)) /=.
  Abort.

  Definition scale_const (coeff : Ri) : polynomial -> polynomial :=
    M.map (fun coeff' => coeff * coeff')%R.
  Definition scale_term (t : term) (p : polynomial) : polynomial :=
    let '(m, coeff) := t in
    let mul1 '(m', coeff') := (m_mul m m', coeff * coeff')%R in
    from_terms (map mul1 (M.elements p)).
  Definition p_add : polynomial -> polynomial -> polynomial :=
    M.map2 (fun mcoeff1 mcoeff2 => match mcoeff1, mcoeff2 with
                                | Some coeff1, Some coeff2 =>
                                  let coeff' := (coeff1 + coeff2)%R in
                                  if coeff' == 0%R then None else Some coeff'
                                | Some coeff1, _ => Some coeff1
                                | _, Some coeff2 => Some coeff2
                                | _, _ => None
                                end).
  Definition p_sub : polynomial -> polynomial -> polynomial :=
    M.map2 (fun mcoeff1 mcoeff2 => match mcoeff1, mcoeff2 with
                                | Some coeff1, Some coeff2 =>
                                  let coeff' := (coeff1 - coeff2)%R in
                                  if coeff' == 0%R then None else Some coeff'
                                | Some coeff1, _ => Some coeff1
                                | _, Some coeff2 => Some (-coeff2)%R
                                | _, _ => None
                                end).
  Definition p_mul (p q : polynomial) : polynomial :=
    foldr p_add zero [seq scale_term t p | t <- terms q].

  Definition var (i : 'I_n) : polynomial := M.add (var_m i) 1%R zero.
  Definition constant (x : Ri) : polynomial :=
    if x == 0%R then zero else M.add m_one x zero.
  Definition eval (p : polynomial) (v : tuple' Ri n) : Ri :=
    foldr (fun x y => x + y)%R 0%R
          [seq let '(m, coeff) := t in coeff * eval_m m v | t <- terms p]%R.

  Definition nconjecture_inst_poly_f : seq polynomial :=
    [seq p_mul (p_sub (var j) (var i))
         (if (nat_of_ord i, nat_of_ord j) != (0, n - 1)
          then foldr p_add zero (map var (irange i j)) else constant 1) |
     i <- seq_iota' n, j <- [seq j <- seq_iota' n | nat_of_ord i < nat_of_ord j]].
  Definition nconjecture_inst_poly : polynomial :=
    foldr p_mul (constant 1) nconjecture_inst_poly_f.

  Definition bounded_product (bound : monomial) : seq polynomial -> polynomial :=
    let mul1 p q := filter_terms (fun '(m, _) => m_cmp m bound) (p_mul p q) in
    foldr mul1 (constant 1).
  Definition nconjecture_inst_poly_BP : polynomial :=
    bounded_product (const_exp_m (n - 1)) nconjecture_inst_poly_f.

  Definition candidate_terms : seq term :=
    [seq t <- terms nconjecture_inst_poly | m_cmp t.1 (const_exp_m (n - 1))].
  Definition candidate_terms' : seq term := terms nconjecture_inst_poly_BP.
End MapPoly.

Module MapPolyArgs3 <: MapPolyArgs.
  Definition n : nat := 3.
  (* TODO use more efficient ints *)
  Definition Ri : idomainType := [idomainType of int].
End MapPolyArgs3.
Module MapPoly3 := MapPoly MapPolyArgs3.

Module MapPolyArgs7 <: MapPolyArgs.
  Definition n : nat := 7.
  (* TODO use more efficient ints *)
  Definition Ri : idomainType := [idomainType of int].
End MapPolyArgs7.
Module MapPoly7 := MapPoly MapPolyArgs7.

Import MapPoly7.

Definition var' (i : nat) : polynomial := var (inord' i).

(*
Time Compute candidate_terms'.
*)
