From Coq Require Import ssreflect ssrfun ssrbool.
Require Import mathcomp.ssreflect.ssrnat.

Require Import mathcomp.ssreflect.bigop.

Require Import mathcomp.ssreflect.eqtype.
Require Import mathcomp.ssreflect.fintype.
Require Import mathcomp.ssreflect.finset.
Require Import mathcomp.ssreflect.seq.
Require Import mathcomp.ssreflect.tuple.

Require Import mathcomp.fingroup.perm.
Require Import mathcomp.algebra.ssralg.
Require Import mathcomp.algebra.zmodp.
Require Import mathcomp.algebra.poly.
Require Import mathcomp.algebra.finalg.
Require Import mathcomp.algebra.ssrint.
Require Import mathcomp.ssreflect.div.
Require Import SsrMultinomials.mpoly.
From mathcomp Require Import order.
From mathcomp Require Export finmap.

Set Bullet Behavior "Strict Subproofs".

Open Scope fset_scope.
Open Scope ring_scope.

Definition partial_sums {R : ringType} {n} (xs : n.-tuple R) : n.+1.-tuple R :=
  [tuple \sum_(j < n | j < i) tnth xs j | i < n.+1].

Section FixedRing.
  Variable R : ringType.

  Definition ndistinct_poly (n : nat) (P : 'I_n -> 'I_n -> bool) : {mpoly R[n]} :=
    \prod_(ij : 'I_n * 'I_n | P ij.1 ij.2) ('X_ij.1 - 'X_ij.2).
  Definition psums_ndistinct_poly_exc {n} (i j : 'I_n) : bool :=
    (((i == 0 :> nat) && (j.+1 == n :> nat)) || (j == i.+1 :> nat))%nat.
  Definition psums_ndistinct_poly (n : nat) : {mpoly R[n]} :=
    ndistinct_poly n.+1 (fun i j => (i < j) && ~~psums_ndistinct_poly_exc i j) \mPo
                   partial_sums [tuple 'X_i | i < n].
  Definition nconjecture_inst_poly (n : nat) : {mpoly R[n]} :=
    ndistinct_poly n (fun i j => i < j) * psums_ndistinct_poly n.
End FixedRing.

Section FixedIdom.
  Variable Ri : idomainType.

  Definition side_conds (r : seq Ri) : bool := (0 \notin r) && (\sum_(x <- r) x != 0).
  Definition conjecture_inst (S : {fset Ri}) : Prop :=
    exists xs : #|`S|.-tuple Ri, perm_eq xs (enum_fset S) && uniq (partial_sums xs).

  Lemma tuple_perm : forall (S : {fset Ri}) (xs : #|`S|.-tuple Ri),
      perm_eq xs S = (seq_fset tt xs `<=` S) && uniq xs.
  Proof.
    move=> S xs.
    have E_size : uniq xs -> #|`seq_fset tt xs| = #|`S|. {
      move=> ?; rewrite size_seq_fset undup_id //; apply: size_tuple.
    }
    apply/idP/andP => [PermEq | [Sub Uniq]].
    - move: (perm_uniq PermEq) (fset_uniq S) => <- Uniq.
      split; last by assumption.
      apply/fsubset_cardP; first by auto.
      move=> x; rewrite seq_fsetE (perm_mem PermEq) //.
    - apply: (uniq_perm Uniq (fset_uniq _)).
      suff : seq_fset tt xs =i S by move=> H x; rewrite -H seq_fsetE //.
      apply/fsubset_cardP; auto.
  Qed.

  Lemma poly_eq (n : nat) (xs : n.-tuple Ri) (i j : 'I_n) :
      tnth xs i == tnth xs j = (('X_i - 'X_j).@[tnth xs] == 0).
  Proof. rewrite GRing.rmorphB /= !mevalXU GRing.subr_eq0 //. Qed.

  Lemma ndistinct_poly_spec : forall (n : nat) (P : 'I_n -> 'I_n -> bool) (xs : n.-tuple Ri),
      (forall i j : 'I_n, P i j -> i != j) ->
      (forall i j : 'I_n, i != j -> P i j || (P j i || (xs`_i != xs`_j))) ->
      uniq xs = ((ndistinct_poly Ri n P).@[tnth xs] != 0).
  Proof.
    move=> n P xs PP1 PP2.
    rewrite /ndistinct_poly GRing.rmorph_prod /=.
    apply/(uniqP 0)/(GRing.prodf_neq0 _ _);
      rewrite size_tuple;
      setoid_rewrite <- poly_eq;
      setoid_rewrite (tnth_nth 0).
    - move=> /= Inj [i' j'] /= /PP1.
      apply/contra => /eqP E.
      apply/eqP/ord_inj/((Inj i' j' (ltn_ord _) (ltn_ord _)) E).
    - move=> /= AllNonzero i j L_i L_j /eqP E.
      pose i' := Ordinal L_i.
      pose j' := Ordinal L_j.
      rewrite (_ : i = i') // (_ : j = j') // in E |- *.
      apply/f_equal/eqP.
      apply: contraLR E => /PP2 /or3P [] // /(AllNonzero (_, _)) //=.
      rewrite eq_sym //.
  Qed.

  Lemma psums_ndistinct_poly_exc_antisym : forall (n : nat) (i j : 'I_n),
      psums_ndistinct_poly_exc i j -> psums_ndistinct_poly_exc j i ->
      i = j.
  Proof.
    move=> n i j /orP [/andP [/eqP E11 /eqP E12] | /eqP E1]
            /orP [/andP [/eqP E21 /eqP E22] | /eqP E2];
            apply: ord_inj; only 1-3: congruence.
    rewrite E1 in E2; case: (Plus.n_SSn _ E2).
  Qed.

  Lemma psums_ndistinct_poly_exc_spec : forall (n : nat) (xs : n.-tuple Ri),
      side_conds xs ->
      forall (i j : 'I_n.+1), i < j -> psums_ndistinct_poly_exc i j ->
                         (partial_sums xs)`_i != (partial_sums xs)`_j.
  Proof.
    move=> n xs /andP [/memPnC Nonzero NonzeroSum] i j L
            /orP [/andP [/eqP E1 /eqP [E2]] | /eqP E].
    - rewrite !nth_mktuple E1 E2 (big_pred0 _ _ _ (tnth xs) ltn0) (big_ord_narrow (leqnn n))
              (eq_bigr (tnth xs)). 2: {
        move=> i' _.
        f_equal; by apply: ord_inj.
      }
      rewrite -(big_tuple _ _ _ xpredT id) eq_sym //.
    - move: (ltn_ord i) (ltn_ord j).
      rewrite !ltnS !nth_mktuple E => LE_i LE_j.
      rewrite (big_ord_narrow LE_i) (big_ord_narrow LE_j) big_ord_recr /=
              [S in _ != S + _](eq_bigr (fun i' => tnth xs (widen_ord LE_i i'))). 2: {
        move=> i' _.
        f_equal; by apply: ord_inj.
      }
      rewrite GRing.addrC -GRing.subr_eq GRing.addrN //.
      apply/Nonzero/mem_tnth.
  Qed.

  Lemma psums_ndistinct_poly_spec : forall (n : nat) (xs : n.-tuple Ri),
      side_conds xs ->
      uniq (partial_sums xs) = ((psums_ndistinct_poly Ri n).@[tnth xs] != 0).
  Proof.
    move=> n xs SideConds.
    rewrite /psums_ndistinct_poly comp_mpoly_meval
            (meval_eq _ (_ : (fun i => _) =1 tnth (partial_sums xs))).
    - apply: ndistinct_poly_spec => i j. (* -[i L_i] [j L_j]. *)
      + rewrite ltn_neqAle => /andP /and_comm [_] /andP [] //.
      + rewrite neq_ltn => /orP L.
        apply/or3P.
        case: L => L; case E1: psums_ndistinct_poly_exc; case E2: psums_ndistinct_poly_exc.
        * constructor 3; by apply: psums_ndistinct_poly_exc_spec.
        * constructor 3; by apply: psums_ndistinct_poly_exc_spec.
        * constructor 1; by apply/andP.
        * constructor 1; by apply/andP.
        * rewrite (psums_ndistinct_poly_exc_antisym _ _ _ E1 E2) ltnn // in L.
        * constructor 2; by apply/andP.
        * constructor 3; rewrite eq_sym; by apply: psums_ndistinct_poly_exc_spec.
        * constructor 2; by apply/andP.
    - move=> i.
      rewrite /partial_sums !tnth_mktuple GRing.rmorph_sum /=.
      apply: eq_bigr => j Lt.
      rewrite tnth_mktuple mevalXU //.
  Qed.

  Theorem nconjecture_inst_poly_spec : forall (S : {fset Ri}),
      side_conds (enum_fset S) ->
      conjecture_inst S <-> exists xs : #|`S|.-tuple Ri,
          (seq_fset tt xs `<=` S) && ((nconjecture_inst_poly Ri #|`S|).@[tnth xs] != 0).
  Proof.
    move=> S Side.
    rewrite /conjecture_inst /nconjecture_inst_poly.
    (* setoid_rewrite (tuple_perm S). *)
    split.
    - move=> [xs /andP [PermEq PsumsUniq]].
      move: (tuple_perm S) {-}PermEq => -> /andP [Sub Uniq].
      exists xs.
      rewrite GRing.rmorphM GRing.mulf_eq0 negb_or -ndistinct_poly_spec.
      + rewrite -psums_ndistinct_poly_spec; first by apply/and3P.
        rewrite /side_conds (perm_mem PermEq) (perm_big _ PermEq) //.
      + move=> ? ?. rewrite ltn_neqAle => /andP [] //.
      + move=> ? ?. rewrite neq_ltn => /orP [] ?; apply/or3P; constructor; assumption.
    - move=> [xs /andP [Sub]].
      rewrite GRing.rmorphM GRing.mulf_eq0 negb_or -ndistinct_poly_spec.
      + move=> /andP [Uniq PsumsUniq].
        move: (conj Sub Uniq) => /andP.
        rewrite -tuple_perm => PermEq.
        rewrite -psums_ndistinct_poly_spec in PsumsUniq; last by
            rewrite /side_conds (perm_mem PermEq) (perm_big _ PermEq) //.
        exists xs.
          by apply/andP.
      + move=> ? ?. rewrite ltn_neqAle => /andP [] //.
      + move=> ? ?. rewrite neq_ltn => /orP [] ?; apply/or3P; constructor; assumption.
  Qed.

  Lemma rem_subset : forall {T : eqType} (x : T) (s : seq T), {subset [predD1 s & x] <= rem x s}.
  Proof.
    move=> T x.
    elim=> [| y ys IH] x' /andP /= [NE In] //.
    move: in_cons In NE => -> /orP [/eqP -> | In] NE; first by rewrite (negPf NE) mem_head.
    case: (@eqP _ y x) => // foo.
      by apply/mem_behead/IH/andP.
  Qed.

  Lemma sum_subset : forall {T : eqType} {F : T -> nat} (s1 s2 : seq T),
      uniq s1 -> {subset s1 <= s2} -> \sum_(x <- s1) F x <= \sum_(x <- s2) F x.
  Proof.
    move=> T F.
    elim => [| x xs IH] s2 Uniq1 Sub.
    - rewrite big_nil //.
    - move: Uniq1 => /andP [/negP NotIn Uniq1].
      rewrite big_cons (big_rem _ (Sub x (mem_head _ _))) /= leq_add2l.
      apply: IH.
      + assumption.
      + move=> x' In.
        apply/rem_subset/andP; split.
        * apply/negP => /eqP Contra.
          rewrite Contra // in In NotIn.
        * apply/Sub/mem_behead/In.
  Qed.

  Lemma div_by_XsubC : forall (n : nat) (i : 'I_n) (c : Ri) (P : {mpoly Ri[n]}),
      exists (Q R : {mpoly Ri[n]}), P = Q * ('X_i - c%:MP) + R /\
                                   {in msupp R, forall (mon : 'X_{1..n}), mon i == 0%nat}.
  Proof.
    move=> n i c P.
    move E: (\sum_(m <- msupp P) m i)%nat => expsum.
    elim/Wf_nat.lt_wf_ind: expsum P E => - [| expsum_pred] IH P /eqP E.
    - exists 0, P; split.
      + rewrite GRing.mul0r GRing.add0r //.
      + move=> /= m /seq_tnthP [m_ix E_ix].
        rewrite big_tnth sum_nat_eq0 in E.
        rewrite -(eqP (implyP (forallP E m_ix) (erefl true))) -E_ix //.
    - have [m0 NonzeroCoeff] : exists2 m0 : 'X_{1..n}, m0 \in msupp P & m0 i != 0%nat. {
        suff [m0_ix NE] : exists m0_ix, tnth (in_tuple (msupp P)) m0_ix i != 0%nat
          by exists (tnth (in_tuple (msupp P)) m0_ix); [apply: mem_tnth | assumption].
        apply/existsP.
        rewrite -negb_forall -(sum_nat_eq0 xpredT)
        -(big_tnth _ _ _ xpredT (fun m : 'X_{1..n} => m i)) (eqP E) //.
      }
      rewrite mcoeff_msupp in NonzeroCoeff.
      move: {2 3}(m0 i) (erefl (m0 i)) => [] // exp_pred E_exp _.
      pose Z := P@_m0 *: 'X_[m0 - U_(i)] * ('X_i - c%:MP).
      have Z_Alt : Z = P@_m0 *: 'X_[m0] - (P@_m0 * c) *: 'X_[m0 - U_(i)]. {
        rewrite /Z GRing.mulrBr -!mpoly_scaleAl
                [_ * c%:MP]GRing.mulrC mul_mpolyC GRing.scalerA
        -mpolyXD (_ :  m0 - U_(i) + U_(i) = m0)%MM // mnmP => j.
        rewrite mnmDE mnmBE mnm1E subnK //.
        case: eqP E_exp => // -> -> //.
      }
      case: (IH _ _ (P - Z) erefl) => [| Q' [R [/eqP IH_E IH_ZeroExp]]]. {
        apply/ltP.
        rewrite Z_Alt GRing.opprB [_ - _]GRing.addrC GRing.addrA -(eqP E)
        -{4}(GRing.addrNK (P@_m0 *: 'X_[m0]) P).
        apply: (leq_ltn_trans (sum_subset _ _ _ (@msuppD_le _ _ _ _))) => //.
        rewrite (perm_big _ (msuppD _)) [msupp (P@_m0 *: _)]msuppMCX //.
        2: {
          move=> m /=.
          rewrite (perm_mem (msupp_rem _ _)) (mem_rem_uniq _ _ m) // mem_seq1.
          apply/negbTE/negP => /andP [/andP [/negP]] //.
        }
        rewrite !big_cat /= ltn_add2l.
        apply: (leq_ltn_trans (sum_subset _ _ _ (@msuppZ_le _ _ _ _))) => //.
        rewrite msuppX !big_seq1 mnmBE E_exp mnm1E eq_refl subn1 //.
      }
      exists (Q' + P@_m0 *: 'X_[m0 - U_(i)]), R; split; last by assumption.
      rewrite GRing.subr_eq GRing.addrAC /Z -GRing.mulrDl in IH_E.
      apply/eqP/IH_E.
  Qed.

  Fixpoint nprod {n} : n.-tuple {fset Ri} -> {fset n.-tuple Ri} :=
    match n with
    | O => fun _ => [fset [tuple]]
    | S n' => fun As => [fset [tuple of x :: (xs : n'.-tuple Ri)]
                    | x in thead As, xs in nprod [tuple of behead As]]
    end.
  Definition nprod' {n} (As : n.-tuple {fset Ri}) : pred (n.-tuple Ri) :=
    fun xs => [forall i, tnth xs i \in tnth As i].
  Lemma nprod_spec : forall (n : nat) (As : n.-tuple {fset Ri}),
      nprod As =i nprod' As.
  Proof.
    elim=> [| n IH] As xs.
    - rewrite /= tuple0 !inE; apply/eqP/forallP => [_ [? ?] |]; auto.
    - change (nprod As) with [fset [tuple of x :: (xs : n.-tuple Ri)]
                             | x in thead As, xs in nprod [tuple of behead As]].
      apply/imfset2P/idP => /= [[x Head [xs' H ->]] | /forallP H].
      + rewrite IH in H; move/forallP in H.
        apply/forallP => -[[| i] L].
        * rewrite [Ordinal _](ord_inj (x2:=ord0)) //.
        * move: (ltnSE L) => L'.
          rewrite [Ordinal _](ord_inj (x2:=inord (Ordinal L').+1)) //;
            last by rewrite /= inordK.
          rewrite -!tnth_behead (tnth_nth 0) /= -(tnth_nth _ _ (Ordinal L')).
          apply: H.
      + exists (thead xs); first by apply: H.
        exists [tuple of behead xs]; last by rewrite {1}[xs]tuple_eta //.
        rewrite IH; apply/forallP => i; rewrite !tnth_behead; apply: H.
  Qed.
  Lemma nprod_spec' : forall (n : nat) (As : n.-tuple {fset Ri}) (xs : n.-tuple Ri),
      xs \in nprod As <-> forall i, tnth xs i \in tnth As i.
  Proof. move=> *; rewrite nprod_spec (rwP forallP) //. Qed.

  Theorem nonvanishing_contra :
    forall (n : nat) (mon : 'X_{1..n}) (P : {mpoly Ri[n]}),
      msize P = (mdeg mon).+1 -> P@_mon != 0 ->
      forall (As : n.-tuple {fset Ri}),
        (forall xs : nprod As, P.@[tnth (val xs)] == 0) ->
        (exists i : 'I_n, #|`tnth As i| <= mon i).
  Proof.
    case. {
      move=> mon P _ NonzeroCoeff As /(_ [`fset11 _]) /= Vanishes.
      suff : P@_mon == 0 by case: (P@_mon == 0) NonzeroCoeff.
      rewrite [x in x.@[_]]nvar0_mpolyC mevalC in Vanishes.
      have -> : mon = 0%MM by apply/mnmP => -[].
      apply: Vanishes => -[] //.
    }
    move=> n' mon.
    move E: (mdeg mon) => deg.
    elim: deg mon E => [| deg_pred IH] mon /eqP MDegree P PDegree NonzeroCoeff As Vanishes.
    - move: (mdeg_eq0 mon) MDegree => -> /eqP -> in NonzeroCoeff |- *.
      move: PDegree => /eqP/msize_poly1P [c Nonzero_c ->] in Vanishes.
      apply/existsP/contraT.
      move: negb_exists => -> /forallP Inhabited.
      suff : c == 0 by case: (c == 0) Nonzero_c.
      set xs := [tuple head 0 (tnth As i) | i < n'.+1].
      suff H : xs \in nprod As by move: mevalC (Vanishes [`H]) => -> //.
      apply/nprod_spec' => i.
      rewrite tnth_mktuple -nth0.
      case E: #|`tnth As i| => [| n].
      + specialize (Inhabited i); rewrite E // in Inhabited.
      + apply: mem_nth; rewrite E //.
    - have [i NonzeroExp] : exists i, mon i > 0. {
        rewrite mdegE in MDegree.
        move/eqP in MDegree.
        have : (deg_pred.+1 != 0)%nat by trivial.
        rewrite -MDegree (sum_nat_eq0 predT mon) negb_forall.
        move=> /existsP [x].
        rewrite negb_imply -lt0n.
        move=> /andP [_ L].
        eauto.
      }
      case: {2}(#|`tnth As i|) (erefl #|`tnth As i|) => [| card_pred] E_Card.
      {by exists i; rewrite E_Card. }
      move: E_Card (ltn0Sn card_pred) => {card_pred} <-.
      rewrite cardfs_gt0 => /fset0Pn /= [a In].
      move: (div_by_XsubC _ i a P) => [Q [R [E ZeroExps_R]]].
      pose As' := [tuple if i == j then tnth As j `\ a else tnth As j | j < n'.+1].
      have E_QCoeff : forall mon', Q@_mon' == P@_(U_(i) + mon') + a * Q@_(U_(i) + mon'). {
        move=> mon'.
        have NonzeroExp' : (U_(i) + mon')%MM i != 0%nat by rewrite mnmDE mnm1E eq_refl //.
        rewrite -GRing.subr_eq E GRing.raddfD /=
                (memN_msupp_eq0 (contra (ZeroExps_R _) NonzeroExp')) GRing.addr0
                GRing.mulrBr GRing.raddfB /= mcoeffMX [Q * a%:MP]GRing.mulrC mcoeffCM //.
      }
      have Q_UBound : msize Q <= deg_pred.+1. {
        rewrite msizeE big_tnth.
        case: {-2}(msupp Q) (erefl (msupp Q)); first by rewrite big_ord0 //.
        move=> m0 ms /= E_m.
        rewrite (bigmax_eq_arg ord0) //.
        case: (arg_maxP _ _) => //= j _.
        set mon' := tnth _ _ => Max.
        have NonzeroCoeff' : mon' \in msupp Q by rewrite -E_m; apply: mem_tnth.
        rewrite mcoeff_msupp (eqP (E_QCoeff mon')) in NonzeroCoeff'.
        case: (@eqP _ P@_(U_(i) + mon')%MM 0) NonzeroCoeff' => [-> | /eqP NE] NonzeroCoeff'.
        - rewrite GRing.add0r GRing.mulf_eq0 negb_or -mcoeff_msupp in NonzeroCoeff'.
          move: NonzeroCoeff' => /andP [_] /seq_tnthP.
          rewrite -E_m => -[j' E_j'].
          move: (Max j' isT).
          rewrite -E_j' mdegD mdeg1 add1n ltnn //.
        - rewrite -mcoeff_msupp in NE.
          move/msize_mdeg_lt in NE.
          rewrite mdegD mdeg1 add1n PDegree ltnS // in NE.
      }
      have ExpandMon : (U_(i) + (mon - U_(i)) = mon)%MM. {
        rewrite mnmP => j.
        rewrite mnmDE mnmBE mnm1E.
        apply: subnKC.
        case: eqP => // <- //.
      }
      move: {-}ExpandMon (E_QCoeff (mon - U_(i))%MM) => -> NonzeroCoeff_Q.
      rewrite (eqP _ : Q@_mon = 0) in NonzeroCoeff_Q. 2: {
        rewrite mcoeff_eq0.
        apply/(contra (@msize_mdeg_lt _ _ _ _)).
        rewrite (eqP MDegree) -leqNgt //.
      }
      rewrite GRing.mulr0 GRing.addr0 in NonzeroCoeff_Q.
      case: (IH (mon - U_(i))%MM _ Q _ _ As' _).
      + apply: eq_add_S.
        rewrite -(eqP MDegree) !mdegE !(big_rem _ (mem_index_enum i)) /= -addSn.
        f_equal.
        * rewrite mnmBE mnm1E eq_refl subn1 prednK //.
        * apply: eq_big_seq => j.
          rewrite mem_rem_uniq; last by rewrite /index_enum -enumT enum_uniq.
          move=> /andP [/negbTE NE _].
          rewrite mnmBE mnm1E eq_sym NE subn0 //.
      + apply/eqP.
        rewrite eqn_leq; apply/andP; split => //.
        rewrite -ltnS -(eqP MDegree) -ExpandMon mdegD mdeg1 add1n ltnS.
        apply: msize_mdeg_lt.
        rewrite mcoeff_msupp (eqP NonzeroCoeff_Q) //.
      + rewrite (eqP NonzeroCoeff_Q) //.
      + move=> [xs /nprod_spec' Ins] /=.
        pose xs' := [tuple if i == j then a else tnth xs j | j < n'.+1].
        have P_Root : P.@[tnth xs] == 0. {
          suff H : xs \in nprod As by apply: (Vanishes [`H]).
          apply/nprod_spec' => j.
          move: tnth_mktuple (Ins j) => ->.
          case: eqP => // <-.
          rewrite in_fsetD1 => /andP [_] //.
        }
        have P_Root' : P.@[tnth xs'] == 0. {
          suff H : xs' \in nprod As by apply: (Vanishes [`H]).
          apply/nprod_spec' => j.
          move: tnth_mktuple (Ins j) => ->.
          rewrite tnth_mktuple.
          case: eqP => // <-.
          rewrite in_fsetD1 => /andP [_] //.
        }
        rewrite E !GRing.rmorphD !GRing.rmorphM !GRing.rmorphB /= !mevalXU !mevalC
          in P_Root P_Root'.
        rewrite tnth_mktuple eq_refl GRing.addrN GRing.mulr0 GRing.add0r in P_Root'.
        suff R_Root : R.@[tnth xs] == 0. {
          rewrite (eqP R_Root) GRing.addr0 GRing.mulf_eq0 in P_Root.
          case: (orP P_Root) (Ins i) => //.
          rewrite GRing.subr_eq0 tnth_mktuple eq_refl in_fsetD1 => E_a /andP [/negP NE_a] //.
        }
        rewrite -(eqP P_Root') !mevalE.
        apply/eqP/eq_big_seq => mon' In_msupp.
        f_equal.
        apply: eq_big => // j _.
        rewrite tnth_mktuple.
        case: (@eqP _ i j) (ZeroExps_R _ In_msupp) => // -> /eqP -> //.
      + move=> j.
        rewrite tnth_mktuple mnmBE mnm1E.
        case: eqP => [E_j | _] /= LE; exists j.
        * rewrite -E_j in LE |- *.
          rewrite -ltnS subn1 (prednK NonzeroExp) in LE.
          rewrite (cardfsD1 a) In //.
        * rewrite subn0 // in LE.
  Qed.

  Corollary nonvanishing :
    forall (n : nat) (mon : 'X_{1..n}) (P : {mpoly Ri[n]}),
      msize P = (mdeg mon).+1 -> P@_mon != 0 ->
      forall (As : n.-tuple {fset Ri}),
        (forall i : 'I_n, #|`tnth As i| > mon i) ->
        exists xs : nprod As, P.@[tnth (val xs)] != 0.
  Proof.
    move=> n mon P E1 E2 As /forallP Exps; apply/existsP; move: Exps.
    apply: contraLR; rewrite negb_exists negb_forall => /forallP Vanishes; apply/existsP.
    setoid_rewrite Bool.negb_involutive in Vanishes; setoid_rewrite <- leqNgt.
      by apply: nonvanishing_contra Vanishes.
  Qed.

  Definition conjecture_at_size (n : nat) : Prop :=
    forall (S : {fset Ri}), #|`S| = n -> side_conds (enum_fset S) -> conjecture_inst S.
  Theorem conjecture_at_size_condition : forall (n : nat),
      let P := nconjecture_inst_poly Ri n in
      (exists mon, mon \in msupp P /\ msize P = (mdeg mon).+1 /\ forall i, mon i < n) ->
      conjecture_at_size n.
  Proof.
    move=> n P [mon [E2 [E1 Exps]]] S ? Side.
    subst.
    rewrite mcoeff_msupp in E2.
    apply/(nconjecture_inst_poly_spec _ Side).
    case: (nonvanishing _ mon P E1 E2 [tuple S | _ < #|`S|] _);
      first by setoid_rewrite tnth_mktuple.
    move=> xs Nonroot; move: (valP xs) => /nprod_spec'; setoid_rewrite tnth_mktuple => ?.
    exists (val xs); apply/andP; split; last by assumption.
    apply/fsubsetP => x; rewrite seq_fsetE => /tnthP [i] -> //.
  Qed.
End FixedIdom.

Definition conjecture : Prop :=
  forall p : nat, prime.prime p -> forall S : {fset 'F_p},
      side_conds [idomainType of 'F_p] (enum S) ->
      conjecture_inst [idomainType of 'F_p] S.

Lemma Z_poly : forall (n : nat) (R : comRingType),
    nconjecture_inst_poly R n = map_mpoly intr (nconjecture_inst_poly [ringType of int] n).
Proof.
  move=> n R.
  rewrite /nconjecture_inst_poly GRing.rmorphM /=
          /psums_ndistinct_poly /ndistinct_poly !GRing.rmorph_prod /=.
  f_equal.
  - apply: eq_bigr => [i j].
    rewrite GRing.rmorphB /= !map_mpolyX //.
  - apply: eq_bigr => -[i j] _ /=.
    rewrite !GRing.rmorphB /= !comp_mpolyXU -!tnth_nth !tnth_mktuple !GRing.rmorph_sum.
    f_equal.
    + apply: eq_bigr => k _ /=.
      rewrite !tnth_mktuple map_mpolyX //.
    + f_equal.
      apply: eq_bigr => k _ /=.
      rewrite !tnth_mktuple map_mpolyX //.
Qed.

Theorem conjecture_at_size_condition_Z : forall (n : nat) (Ri : finIdomainType),
    let P := nconjecture_inst_poly [ringType of int] n in
    (exists mon, intr P@_mon != 0 :> Ri /\ msize P = (mdeg mon).+1 /\ forall i, mon i < n) ->
    conjecture_at_size Ri n.
Proof.
  move=> n Ri P [mon [E2 [E1 Exps]]] S ? Side.
  subst.
  apply: conjecture_at_size_condition => //=.
  rewrite Z_poly -/P.
  exists mon; split; [| split].
  - rewrite mcoeff_msupp mcoeff_map_mpoly //.
  - apply/eqP.
    rewrite eqn_leq.
    apply/andP; split.
    + rewrite -E1 [L in L <= _]msizeE big_tnth.
      apply/bigmax_leqP => /= i _.
      move: (mem_tnth i (in_tuple _)).
      rewrite mcoeff_msupp mcoeff_map_mpoly /= (_ : 0 = 0%:~R) // =>
      /(contra_neq (@f_equal _ _ intr _ _)).
      rewrite -mcoeff_msupp => /msize_mdeg_lt //.
    + rewrite -mcoeff_map_mpoly /= -mcoeff_msupp in E2.
      move: E2 => /msize_mdeg_lt //.
  - assumption.
Qed.

(* why isn't this in mathcomp??? *)
Lemma intr_kernel : forall (R : ringType) (p : nat), p \in [char R] -> forall (n : int),
        (n%:~R == 0 :> R) = (p %| `|n|).
Proof.
  move=> R p Char [] => n; symmetry.
  - rewrite absz_nat.
      by apply/GRing.dvdn_charf.
  - rewrite /= /intr GRing.oppr_eq0.
      by apply/GRing.dvdn_charf.
Qed.

Definition candidate_coeffs (n : nat) : seq int :=
  let P := nconjecture_inst_poly [ringType of int] n in
  [seq P@_mon | mon <- msupp P & ((mdeg mon).+1 == msize P) && [forall i, mon i < n]].

Definition failed_chars (n : nat) : seq nat :=
  prime.primes (foldr gcdn 0%nat (map absz (candidate_coeffs n))).

Theorem failed_chars_spec : forall (n : nat) (Ri : finIdomainType) (p : nat),
    p \in [char Ri] -> candidate_coeffs n != [::] ->
          p \notin failed_chars n -> conjecture_at_size Ri n.
Proof.
  move=> n Ri p Char Nonempty NotFailed.
  apply: conjecture_at_size_condition_Z.
  rewrite /failed_chars foldrE big_tnth prime.mem_primes !negb_and in NotFailed.
  have /hasP /= [m /mapP /= [mon]] : has (fun m => ~~ (p %| `|m|)) (candidate_coeffs n). {
    move: NotFailed => /or3P [].
    - move=> /negP /(_ (GRing.charf_prime Char)) //.
    - case E: candidate_coeffs Nonempty => [| coeff coeffs] //= _.
      rewrite big_ord_recl (tnth_nth 0%nat) /= gcdn_gt0 negb_or -eqn0Ngt absz_eq0 =>
        /andP [/eqP Zero _].
      move: Zero E (mem_head coeff coeffs) => -> <- /mapP /= [mon In Zero'].
      rewrite mem_filter mcoeff_msupp Zero' in In.
      move: In => /andP [_ /eqP] //.
    - rewrite has_predC.
      apply: contra => /allP /= Div.
      apply/dvdn_biggcdP => /= -[j L_j] _.
      rewrite (tnth_nth 0%nat) /=.
      rewrite size_map in L_j.
      rewrite (nth_map 0) //.
        by apply/Div/mem_nth.
  }
  rewrite mem_filter eq_sym -(intr_kernel Ri) // => /andP [/andP [/eqP H2 /forallP H3] _] -> H1.
  eauto.
Qed.

Definition failed_primes (n : nat) : seq nat := filter (ltn n) (failed_chars n).

Theorem failed_primes_spec : forall (n : nat) (p : nat),
    prime.prime p -> candidate_coeffs n != [::] ->
    p \notin failed_primes n -> conjecture_at_size [idomainType of 'F_p] n.
Proof.
  move=> n p Prime Nonempty.
  rewrite /failed_primes mem_filter negb_and /= => /orP [];
    last by apply: (failed_chars_spec _ _ _ (char_Fp _) Nonempty).
  move=> /negP NL S Card /andP [Nonzero _].
  move: (cardfsU1 0 S) => Card'.
  rewrite Nonzero /= add1n Card in Card'.
  have Incl : 0 |` S `<=` [fset x in 'F_p] by apply/fsubsetP => ? _; rewrite !inE.
  move: (fsubset_leq_card Incl).
  rewrite Card' card_finset (card_Fp Prime) //.
Qed.

