{-# LANGUAGE ExistentialQuantification #-}
{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE OverloadedStrings #-}
module DiagramWidgets where

import IHaskell.IPython.Message.UUID
import Diagrams.Prelude hiding ((.=))
import Diagrams.Backend.Cairo
import Data.IORef
import IHaskell.Display
import qualified IHaskell.IPython.Types as IHTypes
import Data.Aeson
import Data.Aeson.Lens
import Diagrams.Backend.Cairo.Internal (Options(..))
import IHaskell.Display.Diagrams
import Data.Maybe
import IHaskell.Eval.Widgets
import Data.String

data DiagramWidget a b =
  Widg {
    wuuid :: UUID,
    wsize :: SizeSpec (V B) (N B),
    wstate :: IORef a,
    wdiacache :: IORef (QDiagram B (V B) (N B) b, Transformation (V B) (N B)),
    wrender :: a -> QDiagram B (V B) (N B) b,
    wreact  :: b -> a -> IO a}

initDiagramWidgetsJS :: DisplayData
initDiagramWidgetsJS = javascript . unlines $ [
  "window.diagram_widgets = {comms: {}};",
  "Jupyter.notebook.kernel.comm_manager.register_target('diagram_widgets',",
  "  (comm, msg) => {",
  "    let uuid = msg.content.comm_id;",
  "    window.diagram_widgets.comms[uuid] = comm;",
  "    // comm.on_msg(msg_ => console.log(msg_));",
  "    comm.on_msg(msg_ => {if (msg_.content.data.action === 'new_val') {",
  "      document.querySelectorAll('[data-widg=\"' + uuid + '\"]').forEach(",
  "        node => node.src = 'data:image/png;base64,' +",
  "          msg_.content.data.new_val)}});",
  "  });",
  "window.diagram_widgets.loadImg = function(uuid) {",
  "  let comm = window.diagram_widgets.comms[uuid],",
  "      msg = {action: 'req_img'};",
  "  comm.send(msg);",
  "}",
  "window.diagram_widgets.clickedImg = function(el, event) {",
  "  let uuid = el.dataset.widg, comm = window.diagram_widgets.comms[uuid],",
  "      x = event.offsetX, y = event.offsetY,",
  "      msg = {action: 'click', loc: {x, y}};",
  "  comm.send(msg);",
  "}"]

transformFor :: Monoid m => SizeSpec (V B) (N B) ->
  QDiagram B (V B) (N B) m -> Transformation (V B) (N B)
transformFor spec dia = t
  where (_, t, _) = adjustDia Cairo (CairoOptions "" spec RenderOnly False) dia

instance Monoid b => IHaskellDisplay (DiagramWidget a b) where
  display Widg{wuuid} =
    return . Display . pure . html . concat $ [
      "<div>",
      "<img data-widg=\"", suuid,
      "\" onclick=\"diagram_widgets.clickedImg(this, event)\">",
      "<script>diagram_widgets.loadImg('", suuid, "')</script>",
      "</div>"]
    where suuid = uuidToString wuuid

instance Monoid b => IHaskellWidget (DiagramWidget a b) where
  getCommUUID Widg{wuuid} = wuuid
  targetName _ = "diagram_widgets"
  comm Widg{wsize, wstate, wdiacache, wrender, wreact} v reply = do
    case v ^? _Object.ix "action"._String of
      Just "click" -> case (v ^? _Object.ix "loc"._Object.ix "x"._Double,
                            v ^? _Object.ix "loc"._Object.ix "y"._Double) of
        (Just x, Just y) -> do
          (dia, t) <- readIORef wdiacache
          let b = sample dia (papply (inv t) (p2 (x, y)))
          a <- readIORef wstate
          a' <- wreact b a
          writeIORef wstate a'
          let dia' = wrender a'
              t' = transformFor wsize dia'
          writeIORef wdiacache (dia', t')
        _ -> return ()
      _ -> return ()
    (dia, _) <- readIORef wdiacache
    Display opts <- display (withSizeSpec wsize (clearValue dia))
    let grabPNG dd = do DisplayData (IHTypes.MimePng _ _) d <- Just dd; pure d
        [dat] = mapMaybe grabPNG opts
    reply (object ["action" .= String "new_val", "new_val" .= dat])

mkDiagramWidget :: Monoid b =>
  a -> SizeSpec (V B) (N B) -> (a -> QDiagram B (V B) (N B) b) ->
  (b -> a -> IO a) -> IO (DiagramWidget a b)
mkDiagramWidget a wsize wrender wreact = do
  wuuid <- IHaskell.IPython.Message.UUID.random
  wstate <- newIORef a
  let dia = wrender a
      t = transformFor wsize dia
  wdiacache <- newIORef (dia, t)
  let w = Widg{wuuid, wsize, wstate, wdiacache, wrender, wreact}
  widgetSendOpen w (fromString "placeholder")
  return w


-- e.g.

sampleDiagramWidget :: IO (DiagramWidget Int Any)
sampleDiagramWidget =
  mkDiagramWidget 10 (dims2D 300 300)
  (\n -> text (show n) <> circle 10 # fc gray)
  (\(Any b) n -> return $ if b then n + 1 else n - 1)

